from typing import Any, Tuple

import os

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class OsStatResultSerializer(SerializerInterface):
    """
    Serialize and deserialize `os.stat_result` into `PJSONType`.

    `os.stat_result` is basically a wrapper around a tuple of ints, so we can serialize it as that tuple.
    """

    targets = (os.stat_result,)

    def obj_to_pjson(self, obj: os.stat_result, type_hint: Any) -> PJSONType:
        return self._service.to_pjson(obj, Tuple[int, ...])

    def pjson_to_obj(self, pjson_obj: PJSONType, type_hint: Any) -> os.stat_result:
        return os.stat_result(pjson_obj)  # type: ignore



import asyncio
import dataclasses
import hashlib
import logging
from inspect import isawaitable
from typing import (
    BinaryIO,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
    Union,
    Awaitable,
    Sequence,
    Callable,
    Set,
)

from ofrak.component.interface import ComponentInterface
from ofrak.model.component_model import ComponentContext, CC, ComponentRunResult
from ofrak.model.data_model import DataPatch
from ofrak.model.job_model import (
    JobRunContext,
)
from ofrak.model.job_request_model import (
    JobAnalyzerRequest,
    JobComponentRequest,
    JobMultiComponentRequest,
)
from ofrak.model.resource_model import (
    ResourceAttributes,
    ResourceModel,
    MutableResourceModel,
    ResourceContext,
    Data,
)
from ofrak.model.tag_model import ResourceTag
from ofrak.model.viewable_tag_model import (
    ViewableResourceTag,
    ResourceViewInterface,
    ResourceViewContext,
)
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.dependency_handler import DependencyHandler
from ofrak.service.id_service_i import IDServiceInterface
from ofrak.service.job_service_i import JobServiceInterface
from ofrak.service.resource_service_i import (
    ResourceServiceInterface,
    ResourceFilter,
    ResourceSort,
)
from ofrak_type.error import NotFoundError, InvalidStateError
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)
RT = TypeVar("RT", bound="ResourceTag")
RA = TypeVar("RA", bound="ResourceAttributes")
RV = TypeVar("RV", bound="ResourceViewInterface")


class Resource:
    """
    Defines methods for interacting with the data and attributes of Resources, the main building
    block of OFRAK.
    """

    __slots__ = (
        "_job_id",
        "_job_context",
        "_component_context",
        "_resource_context",
        "_resource_view_context",
        "_resource",
        "_resource_factory",
        "_id_service",
        "_resource_service",
        "_data_service",
        "_job_service",
        "_dependency_handler",
    )

    def __init__(
        self,
        job_id: bytes,
        resource: MutableResourceModel,
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        job_context: Optional[JobRunContext],
        component_context: ComponentContext,
        resource_factory: "ResourceFactory",
        id_service: IDServiceInterface,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        job_service: JobServiceInterface,
    ):
        self._job_id: bytes = job_id
        self._job_context: Optional[JobRunContext] = job_context
        self._component_context: ComponentContext = component_context
        self._resource_context: ResourceContext = resource_context
        self._resource_view_context: ResourceViewContext = resource_view_context
        self._resource: MutableResourceModel = resource

        self._resource_factory: "ResourceFactory" = resource_factory
        self._id_service: IDServiceInterface = id_service
        self._resource_service: ResourceServiceInterface = resource_service
        self._data_service: DataServiceInterface = data_service
        self._job_service: JobServiceInterface = job_service

    def get_id(self) -> bytes:
        """
        :return: This resource's ID
        """
        return self._resource.id

    def get_job_id(self) -> bytes:
        """
        Each resource belongs to a specific "job." See
        [JobServiceInterface][ofrak.service.job_service_i.JobServiceInterface].

        :return: The ID of the job this resource belongs to
        """
        return self._job_id

    def get_data_id(self) -> Optional[bytes]:
        """
        Each resource may have a data ID. This refers to a
        [DataModel][ofrak.model.data_model.DataModel] representing some chunk of raw binary data.

        :return: The data ID associated with this resource, if it exists
        """
        return self._resource.data_id

    def get_resource_context(self) -> ResourceContext:
        return self._resource_context

    def get_resource_view_context(self) -> ResourceViewContext:
        return self._resource_view_context

    def get_component_context(self) -> ComponentContext:
        return self._component_context

    def get_job_context(self) -> Optional[JobRunContext]:
        return self._job_context

    def get_caption(self) -> str:
        return self._resource.caption

    def is_modified(self) -> bool:
        """
        Check if the resource has been modified in this context and is considered "dirty".
        :return: `True` if the resource is modified, `False` otherwise
        """
        return self._resource.is_modified

    def get_model(self) -> MutableResourceModel:
        """
        Get the underlying [model][ofrak.model.resource_model.ResourceModel] of this resource.
        :return:
        """
        return self._resource

    async def get_data(self, range: Optional[Range] = None) -> bytes:
        """
        A resource often represents a chunk of underlying binary data. This method returns the
        entire chunk by default; this can be reduced by an optional parameter.

        :param range: A range within the resource's data, relative to the resource's data itself
        (e.g. Range(0, 10) returns the first 10 bytes of the chunk)

        :return: The full range or a partial range of this resource's bytes
        """
        if self._resource.data_id is None:
            raise ValueError(
                "Resource does not have a data_id. Cannot get data from a resource with no data"
            )
        data = await self._data_service.get_data(self._resource.data_id, range)
        if range is None:
            range = Range(0, len(data))
        self._component_context.access_trackers[self._resource.id].data_accessed.add(range)
        return data

    async def get_data_length(self) -> int:
        """
        :return: The length of the underlying binary data this resource represents
        """
        if self._resource.data_id is None:
            raise ValueError(
                "Resource does not have a data_id. Cannot get data length from a "
                "resource with no data."
            )
        return await self._data_service.get_data_length(self._resource.data_id)

    async def get_data_range_within_parent(self) -> Range:
        """
        If this resource is "mapped," i.e. its underlying data is defined as a range of its parent's
        underlying data, this method returns the range within the parent resource's data where this
        resource lies. If this resource is not mapped (it is root), it returns a range starting at 0
        with length 0.

        :return: The range of the parent's data which this resource represents
        """
        if self._resource.data_id is None:
            raise ValueError(
                "Resource does not have a data_id. Cannot get data range from a "
                "resource with no data."
            )
        if self._resource.parent_id is None:
            return Range(0, 0)

        parent_models = list(
            await self._resource_service.get_ancestors_by_id(self._resource.id, max_count=1)
        )
        if len(parent_models) != 1:
            raise NotFoundError(f"There is no parent for resource {self._resource.id.hex()}")
        parent_model = parent_models[0]

        parent_data_id = parent_model.data_id
        if parent_data_id is None:
            return Range(0, 0)

        try:
            return await self._data_service.get_range_within_other(
                self._resource.data_id, parent_data_id
            )
        except ValueError:
            return Range(0, 0)

    async def get_data_range_within_root(self) -> Range:
        """
        Does the same thing as `get_data_range_within_parent`, except the range is relative to the
        root.

        :return: The range of the root node's data which this resource represents
        """
        if self._resource.data_id is None:
            raise ValueError(
                "Resource does not have a data_id. Cannot get data range from a "
                "resource with no data."
            )
        return await self._data_service.get_data_range_within_root(self._resource.data_id)

    async def save(self):
        """
        If this resource has been modified, update the model stored in the resource service with
        the local changes.

        :raises NotFoundError: If the resource service does not have a model for this resource's ID
        """
        await save_resources(
            (self,),
            self._resource_service,
            self._data_service,
            self._component_context,
            self._resource_context,
            self._resource_view_context,
        )

    def _save(
        self,
        resources_to_delete: List[bytes],
        patches_to_apply: List[DataPatch],
        resources_to_update: List[MutableResourceModel],
    ):
        if self._resource.is_deleted:
            resources_to_delete.append(self._resource.id)
        elif self._resource.is_modified:
            modification_tracker = self._component_context.modification_trackers.get(
                self._resource.id
            )
            assert modification_tracker is not None, (
                f"Resource {self._resource.id.hex()} was "
                f"marked as modified but is missing a tracker!"
            )
            patches_to_apply.extend(modification_tracker.data_patches)
            resources_to_update.append(self._resource)
            modification_tracker.data_patches.clear()

    async def _fetch(self, resource: MutableResourceModel):
        """
        Update the local model with the latest version from the resource service. This will fail
        if this resource has been modified.

        :raises InvalidStateError: If the local resource model has been modified
        :raises NotFoundError: If the resource service does not have a model for this resource's ID
        """
        if resource.is_modified and not resource.is_deleted:
            raise InvalidStateError(
                f"Cannot fetch dirty resource {resource.id.hex()} (resource "
                f"{self.get_id().hex()} attempted fetch)"
            )
        try:
            fetched_resource = await self._resource_service.get_by_id(resource.id)
        except NotFoundError:
            if resource.id in self._component_context.modification_trackers:
                del self._resource_context.resource_models[resource.id]
            return

        resource.reset(fetched_resource)

    async def _fetch_resources(self, resource_ids: Iterable[bytes]):
        tasks = []
        for resource_id in resource_ids:
            context_resource = self._resource_context.resource_models.get(resource_id)
            if context_resource is not None:
                tasks.append(self._fetch(context_resource))
        await asyncio.gather(*tasks)

    async def _update_views(self, modified: Set[bytes], deleted: Set[bytes]):
        for resource_id in modified:
            views_in_context = self._resource_view_context.views_by_resource[resource_id]
            for view in views_in_context.values():
                if resource_id not in self._resource_context.resource_models:
                    await self._fetch(view.resource.get_model())  # type: ignore
                updated_model = self._resource_context.resource_models[resource_id]
                fresh_view = view.create(updated_model)
                for field in dataclasses.fields(fresh_view):
                    if field.name == "_resource":
                        continue
                    setattr(view, field.name, getattr(fresh_view, field.name))

        for resource_id in deleted:
            views_in_context = self._resource_view_context.views_by_resource[resource_id]
            for view in views_in_context.values():
                view.set_deleted()

    async def run(
        self,
        component_type: Type[ComponentInterface[CC]],
        config: CC = None,
    ) -> ComponentRunResult:
        """
        Run a single component. Runs even if the component has already been run on this resource.

        :param component_type: The component type (may be an interface) to get and run
        :param config: Optional config to pass to the component

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        job_context = self._job_context
        component_result = await self._job_service.run_component(
            JobComponentRequest(
                self._job_id,
                self._resource.id,
                component_type.get_id(),
                config,
            ),
            job_context,
        )
        for deleted_id in component_result.resources_deleted:
            if deleted_id in self._component_context.modification_trackers:
                del self._component_context.modification_trackers[deleted_id]
        await self._fetch_resources(component_result.resources_modified)
        await self._update_views(
            component_result.resources_modified, component_result.resources_deleted
        )
        return component_result

    async def auto_run(
        self,
        components: Iterable[Type[ComponentInterface]] = tuple(),
        blacklisted_components: Iterable[Type[ComponentInterface]] = tuple(),
        all_unpackers: bool = False,
        all_identifiers: bool = False,
        all_analyzers: bool = False,
        all_packers: bool = False,
    ) -> ComponentRunResult:
        """
        Automatically run multiple components which may run on this resource. From an initial set
        of possible components to run, this set is searched for components for which the
        intersection of the component's targets and this resource's tags is not empty. Accepts
        several optional flags to expand or restrict the initial set of components.

        :param components: Components to explicitly add to the initial set of components
        :param blacklisted_components: Components to explicitly remove to the initial set of
        components
        :param all_unpackers: If true, all Unpackers are added to the initial set of components
        :param all_identifiers: If true, all Identifiers are added to the initial set of components
        :param all_analyzers: If true, all Analyzers are added to the initial set of components

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        components_result = await self._job_service.run_components(
            JobMultiComponentRequest(
                self._job_id,
                self._resource.id,
                components_allowed=tuple(c.get_id() for c in components),
                components_disallowed=tuple(c.get_id() for c in blacklisted_components),
                all_unpackers=all_unpackers,
                all_identifiers=all_identifiers,
                all_analyzers=all_analyzers,
                all_packers=all_packers,
            )
        )
        for deleted_id in components_result.resources_deleted:
            if deleted_id in self._component_context.modification_trackers:
                del self._component_context.modification_trackers[deleted_id]
        await self._fetch_resources(components_result.resources_modified)
        await self._update_views(
            components_result.resources_modified, components_result.resources_deleted
        )
        return components_result

    async def unpack(self) -> ComponentRunResult:
        """
        Unpack the resource.

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        return await self.auto_run(all_identifiers=True, all_unpackers=True)

    async def analyze(self, resource_attributes: Type[RA]) -> RA:
        """
        Analyze the resource for a specific resource attribute.

        :param Type[RA] resource_attributes:

        :return:
        """
        attributes = self._check_attributes(resource_attributes)
        if attributes is None:
            await self._analyze_attributes(resource_attributes)
            return self.get_attributes(resource_attributes)
        else:
            return attributes

    async def identify(self) -> ComponentRunResult:
        """
        Run all registered identifiers on the resource, tagging it with matching resource tags.
        """
        return await self.auto_run(all_identifiers=True)

    async def pack(self) -> ComponentRunResult:
        """
        Pack the resource.

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        return await self.auto_run(all_packers=True)

    async def auto_run_recursively(
        self,
        components: Iterable[Type[ComponentInterface]] = tuple(),
        blacklisted_components: Iterable[Type[ComponentInterface]] = tuple(),
        blacklisted_tags: Iterable[ResourceTag] = tuple(),
        all_unpackers: bool = False,
        all_identifiers: bool = False,
        all_analyzers: bool = False,
    ) -> ComponentRunResult:
        """
        Automatically run multiple components which may run on this resource or its descendents.
        From an initial set of possible components to run, this set is searched for components
        for which the intersection of the component's targets and this resource's tags is not
        empty. Accepts several optional flags to expand or restrict the initial set of
        components.
        After each run, compatible components from the initial set are run on any resources which
        have had tags added (including newly created resources). This is repeated until no new
        tags are added.

        :param components: Components to explicitly add to the initial set of components
        :param blacklisted_components: Components to explicitly remove to the initial set of
        components
        :param all_unpackers: If true, all Unpackers are added to the initial set of components
        :param all_identifiers: If true, all Identifiers are added to the initial set of components
        :param all_analyzers: If true, all Analyzers are added to the initial set of components

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        components_result = await self._job_service.run_components_recursively(
            JobMultiComponentRequest(
                self._job_id,
                self._resource.id,
                components_allowed=tuple(c.get_id() for c in components),
                components_disallowed=tuple(c.get_id() for c in blacklisted_components),
                all_unpackers=all_unpackers,
                all_identifiers=all_identifiers,
                all_analyzers=all_analyzers,
                tags_ignored=tuple(blacklisted_tags),
            )
        )
        await self._fetch_resources(components_result.resources_modified)
        await self._update_views(
            components_result.resources_modified, components_result.resources_deleted
        )
        return components_result

    async def unpack_recursively(
        self,
        blacklisted_components: Iterable[Type[ComponentInterface]] = tuple(),
        do_not_unpack: Iterable[ResourceTag] = tuple(),
    ) -> ComponentRunResult:
        """
        Automatically unpack this resource and recursively unpack all of its descendants. First
        this resource is unpacked; then, any resource which "valid" tags were added to will also be
        unpacked. New resources created with tags count as resources with new tags. A "valid" tag
        is a tag which is not explicitly ignored via the ``do_not_unpack`` argument.
        The unpacking will only stop when no new "valid" tags have been added in the previous
        iteration. This can lead to a very long unpacking process if it is totally unconstrained.

        :param blacklisted_components: Components which are blocked from running during the
        recursive unpacking, on this resource or any descendants.
        :param do_not_unpack: Do not unpack resources with this tag, and ignore these tags when
        checking if any new tags have been added in this iteration.

        :return: A ComponentRunResult containing information on resources affected by the component
        """
        return await self.auto_run_recursively(
            all_identifiers=True,
            all_unpackers=True,
            blacklisted_components=blacklisted_components,
            blacklisted_tags=do_not_unpack,
        )

    async def analyze_recursively(self) -> ComponentRunResult:
        return await self.auto_run_recursively(all_analyzers=True)

    async def pack_recursively(self) -> ComponentRunResult:
        """
        Recursively pack the resource, starting with its descendants.
        """
        return await self._job_service.pack_recursively(self._job_id, self._resource.id)

    async def write_to(self, destination: BinaryIO, pack: bool = True):
        """
        Recursively repack resource and write data out to an arbitrary ``BinaryIO`` destination.
        :param destination: Destination for packed resource data
        :return:
        """
        if pack is True:
            await self.pack_recursively()

        destination.write(await self.get_data())

    async def _analyze_attributes(self, attribute_type: Type[ResourceAttributes]):
        job_context = self._job_context
        components_result = await self._job_service.run_analyzer_by_attribute(
            JobAnalyzerRequest(
                self._job_id,
                self._resource.id,
                attribute_type,
                tuple(self._resource.tags),
            ),
            job_context,
        )
        # Update all the resources in the local context that were modified as part of the
        # analysis
        await self._fetch_resources(components_result.resources_modified)
        await self._update_views(
            components_result.resources_modified, components_result.resources_deleted
        )
        return components_result

    async def _create_resource(self, resource_model: ResourceModel) -> "Resource":
        return await self._resource_factory.create(
            self._job_id,
            resource_model.id,
            self._resource_context,
            self._resource_view_context,
            self._component_context,
            self._job_context,
        )

    async def _create_resources(
        self, resource_models: Iterable[ResourceModel]
    ) -> Iterable["Resource"]:
        return await self._resource_factory.create_many(
            self._job_id,
            [resource_model.id for resource_model in resource_models],
            self._resource_context,
            self._resource_view_context,
            self._component_context,
            self._job_context,
        )

    async def create_child(
        self,
        tags: Iterable[ResourceTag] = None,
        attributes: Iterable[ResourceAttributes] = None,
        data: Optional[bytes] = None,
        data_range: Optional[Range] = None,
    ) -> "Resource":
        """
        Create a new resource as a child of this resource. This method entirely defines the
        child's tags and attributes. This method also defines the child's data semantics:

        A child resource can either be defined in one of three ways:
        1) The resource contains no data ("Dataless" resource). Not used in practice.
        2) As mapping a range of its parent's data ("Mapped" resource). For example, an instruction
        maps a portion of its parent basic block.
        3) Defining its own new, independent data ("Unmapped" resource). For example,
        a file extracted from a zip archive is a child of the zip archive resource, but its data
        does not map to some specific range of that parent archive.

        By default a resource will be defined the third way (unmapped). To specify that the
        resource is a mapped resource, include the optional ``data_range`` parameter set to the
        range of the parent's data which the child maps. That is, `data_range=Range(0,
        10)` creates a resource which maps the first 10 bytes of the parent.
        The optional ``data`` param defines whether to populate the new child's data. It can be used
        only if the data is unmapped. If the child is unmapped, the value of ``data``
        still becomes that child's data, but the parent's data is unaffected. If ``data`` and
        ``data_range`` are both `None` (default), the new child is a dataless resource.

        The following table sums up the possible interactions between ``data`` and ``data_range``:

        |                          | ``data_range`` param not `None`                        | ``data_range`` param `None`                  |
        |--------------------------|--------------------------------------------------------|----------------------------------------------|
        | ``data`` param not `None` | Not allowed                                            | Child unmapped, child's data set to ``data`` |
        | ``data`` param   `None`   | Child mapped, parent's data untouched                  | Child is dataless                            |

        :param tags: [tags][ofrak.model.tag_model.ResourceTag] to add to the new child
        :param attributes: [attributes][ofrak.model.resource_model.ResourceAttributes] to add to
        the new child
        :param data: The binary data for the new child. If `None` and ``data_range`` is `None`,
        the resource has no data. Defaults to `None`.
        :param data_range: The range of the parent's data which the new child maps. If `None` (
        default), the child will not map the parent's data.
        :return:
        """
        if data is not None and data_range is not None:
            raise ValueError(
                "Cannot create a child from both data and data_range. These parameters are "
                "mutually exclusive."
            )
        resource_id = self._id_service.generate_id()
        if data_range is not None:
            if self._resource.data_id is None:
                raise ValueError(
                    "Cannot create a child with mapped data from a parent that doesn't have data"
                )
            data_model_id = resource_id
            await self._data_service.create_mapped(
                data_model_id,
                self._resource.data_id,
                data_range,
            )
            data_attrs = Data(data_range.start, data_range.length())
            attributes = [data_attrs, *attributes] if attributes else [data_attrs]
        elif data is not None:
            if self._resource.data_id is None:
                raise ValueError(
                    "Cannot create a child with data from a parent that doesn't have data"
                )
            data_model_id = resource_id
            await self._data_service.create_root(data_model_id, data)
            data_attrs = Data(0, len(data))
            attributes = [data_attrs, *attributes] if attributes else [data_attrs]
        else:
            data_model_id = None
        resource_model = ResourceModel.create(
            resource_id,
            data_model_id,
            self._resource.id,
            tags,
            attributes,
            self._component_context.component_id,
            self._component_context.component_version,
        )
        await self._resource_service.create(resource_model)
        if self._job_context:
            resource_tracker = self._job_context.trackers[resource_model.id]
            resource_tracker.tags_added.update(resource_model.tags)
        self._component_context.mark_resource_modified(resource_id)
        self._component_context.resources_created.add(resource_model.id)
        created_resource = await self._create_resource(resource_model)
        return created_resource

    async def create_child_from_view(
        self,
        view: RV,
        data_range: Optional[Range] = None,
        data: Optional[bytes] = None,
        additional_tags: Iterable[ResourceTag] = (),
        additional_attributes: Iterable[ResourceAttributes] = (),
    ) -> "Resource":
        """
        Create a new resource as a child of this resource. The new resource will have tags and
        attributes as defined by the [view][ofrak.model.viewable_tag_model.ViewableResourceTag];
        in this way a view can act as a template to create a new resource.

        The ``additional_tags`` and ``additional_attributes`` can also be used to add more tags
        and attributes beyond what the view contains.

        This method's ``data`` and ``data_range`` parameters have the same semantics as in
        `create_child`, in short:

        |                          | ``data_range`` param not `None`                        | ``data_range`` param `None`                  |
        |--------------------------|--------------------------------------------------------|----------------------------------------------|
        | ``data`` param not `None` | Child mapped, ``data`` patched into child (and parent) | Child unmapped, child's data set to ``data`` |
        | ``data`` param   `None`   | Child mapped, parent's data untouched                  | Child is dataless                            |

        See `create_child` documentation for details.

        :param view: A [resource view][ofrak.resource_view] to pull
        [tags][ofrak.model.tag_model.ResourceTag] and
        [attributes][ofrak.model.resource_model.ResourceAttributes] from to populate the new child
        :param data_range: The range of the parent's data which the new child maps. If `None` (
        default), the child will not map the parent's data.
        :param data: The binary data for the new child. If `None` and ``data_range`` is `None`,
        the resource has no data. Defaults to `None`.
        :param additional_tags: Any [tags][ofrak.model.tag_model.ResourceTag] for the child in
        addition to those from the ``view``
        :param additional_attributes: Any
        [attributes][ofrak.model.resource_model.ResourceAttributes] for the child in addition to
        those from the ``view``
        :return:
        """
        viewable_tag: ViewableResourceTag = type(view)
        new_resource = await self.create_child(
            tags=(viewable_tag, *additional_tags),
            attributes=(*view.get_attributes_instances().values(), *additional_attributes),
            data_range=data_range,
            data=data,
        )
        return new_resource

    def _view_as(self, viewable_tag: Type[RV]) -> Union[RV, Awaitable[RV]]:
        """
        Try to get a view without calling any analysis, to avoid as many unnecessary
        `asyncio.gather` calls as possible.

        Checks cached views first for view, and if not found, then checks if the attributes needed
        to create the view are already present and up-to-date, and only if both of those are not
        found does it return an awaitable.
        """
        if self._resource_view_context.has_view(self.get_id(), viewable_tag):
            # First early return: View already exists in cache
            return self._resource_view_context.get_view(self.get_id(), viewable_tag)
        if not issubclass(viewable_tag, ResourceViewInterface):
            raise ValueError(
                f"Cannot get view for resource {self.get_id().hex()} of a type "
                f"{viewable_tag.__name__} because it is not a subclass of ResourceView"
            )
        if not self.has_tag(viewable_tag):
            raise ValueError(
                f"Cannot get resource {self.get_id().hex()} as view "
                f"{viewable_tag.__name__} because the resource is not tagged as a "
                f"{viewable_tag.__name__}"
            )
        composed_attrs_types = viewable_tag.composed_attributes_types
        existing_attributes = [self._check_attributes(attrs_t) for attrs_t in composed_attrs_types]
        if all(existing_attributes):
            # Second early return: All attributes needed for view are present and up-to-date
            view = viewable_tag.create(self.get_model())
            view.resource = self  # type: ignore
            self._resource_view_context.add_view(self.get_id(), view)
            return cast(RV, view)

        analysis_tasks = [
            self._analyze_attributes(attrs_t)
            for attrs_t, existing in zip(composed_attrs_types, existing_attributes)
            if not existing
        ]

        # Only if analysis is absolutely necessary is an awaitable created and returned
        async def finish_view_creation() -> RV:
            await asyncio.gather(*analysis_tasks)
            view = viewable_tag.create(self.get_model())
            view.resource = self  # type: ignore
            self._resource_view_context.add_view(self.get_id(), view)
            return cast(RV, view)

        return finish_view_creation()

    async def view_as(self, viewable_tag: Type[RV]) -> RV:
        """
        Provides a specific type of view instance for this resource. The returned instance is an
        object which has some of the information from this same resource, however in a simpler
        interface. This resource instance will itself remain available through the view's
        ``.resource`` property.
        :param viewable_tag: A ViewableResourceTag, which this resource's model must already contain

        :raises ValueError: If the model does not contain this tag, or this tag is not a
        ViewableResourceTag

        :return:
        """
        view_or_create_view_task: Union[RV, Awaitable[RV]] = self._view_as(viewable_tag)
        if isawaitable(view_or_create_view_task):
            return await view_or_create_view_task
        else:
            return cast(RV, view_or_create_view_task)

    def add_view(self, view: ResourceViewInterface):
        """
        Add all the attributes composed in a view to this resource, and tag this resource with
        the view type. Calling this is the equivalent of making N ``add_attributes`` calls and
        one ``add_tag`` call (where N is the number of attributes the view is composed of).

        :param view: An instance of a view
        """
        for attributes in view.get_attributes_instances().values():  # type: ignore
            self.add_attributes(attributes)
        self.add_tag(type(view))

    def _set_modified(self):
        self._component_context.mark_resource_modified(self._resource.id)

    def _add_tag(self, tag: ResourceTag):
        """
        Associate a tag with the resource. If the resource already have the provided tag, it has no
        effects. All parent classes of the provided tag that are tags themselves are also added.
        """
        if self._resource.has_tag(tag, False):
            return
        self._component_context.mark_resource_modified(self._resource.id)
        new_tags = self._resource.add_tag(tag)
        if self._job_context:
            resource_tracker = self._job_context.trackers[self._resource.id]
            resource_tracker.tags_added.update(new_tags)

    def add_tag(self, *tags: ResourceTag):
        """
        Associate multiple tags with the resource. If the resource already have one of the provided
        tag, the tag is not added. All parent classes of the provided tag that are tags themselves
        are also added.
        """
        for tag in tags:
            self._add_tag(tag)

    def get_tags(self, inherit: bool = True) -> Iterable[ResourceTag]:
        """
        Get a set of tags associated with the resource.
        """
        return self._resource.get_tags(inherit)

    def has_tag(self, tag: ResourceTag, inherit: bool = True) -> bool:
        """
        Determine if the resource is associated with the provided tag.
        """
        return self._resource.has_tag(tag, inherit)

    def remove_tag(self, tag: ResourceTag):
        if not self._resource.has_tag(tag):
            return
        self._set_modified()
        self._resource.remove_tag(tag)

    def get_most_specific_tags(self) -> Iterable[ResourceTag]:
        """
        Get all tags associated with the resource from which no other tags on that resource
        inherit. In other words, get the resource's tags that aren't subclassed by other tags on
        the resource.

        For example, for a resource tagged as `Elf`, the result would be just `[Elf]` instead of
        `[Elf, Program, GenericBinary]` that `Resource.get_tags` returns. This is because `Elf`
        inherits from `Program`, which inherits from `GenericBinary`. Even though the resource
        has all of those tags, the most derived class with no other derivatives is the "most
        specific."
        """
        return self._resource.get_most_specific_tags()

    def _check_attributes(self, attributes_type: Type[RA]) -> Optional[RA]:
        """
        Try to get the current attributes.

        TODO: Should we be using the version as well? The client wouldn't know the
        version of the component in a client-server environment. We could do that efficiently by
        adding a service method that list all available components (and their version)

        :param attributes_type: The type of attributes to check this resource for.

        :return: The requested attributes if they are present and up-to-date, otherwise return None.
        """
        attributes = self._resource.get_attributes(attributes_type)
        if attributes is not None:
            # Make sure that the attributes have not been invalidated
            component_id = self._resource.get_component_id_by_attributes(type(attributes))
            if component_id is not None:
                return attributes
        return None

    def _add_attributes(self, attributes: ResourceAttributes):
        existing_attributes = self._resource.get_attributes(type(attributes))
        if existing_attributes is not None and existing_attributes == attributes:
            return
        self._set_modified()
        self._resource.add_attributes(attributes)
        component_context = self._component_context
        self._resource.add_component_for_attributes(
            component_context.component_id, component_context.component_version, type(attributes)
        )

    def add_attributes(self, *attributes: ResourceAttributes):
        """
        Add the provided attributes to the resource. If the resource already have the
        provided attributes classes, they are replaced with the provided one.
        """
        for attrs in attributes:
            self._add_attributes(attrs)

    def has_attributes(self, attributes_type: Type[ResourceAttributes]) -> bool:
        """
        Check if this resource has a value for the given attributes type.
        :param attributes_type:
        :return:
        """
        return self._resource.has_attributes(attributes_type)

    def get_attributes(self, attributes_type: Type[RA]) -> RA:
        """
        If this resource has attributes matching the given type, return the value of those
        attributes. Otherwise returns `None`.
        :param attributes_type:
        :return:
        """
        attributes = self._resource.get_attributes(attributes_type)
        if attributes is None:
            raise NotFoundError(
                f"Cannot find attributes {attributes_type} for resource {self.get_id().hex()}"
            )

        self._component_context.access_trackers[self._resource.id].attributes_accessed.add(
            attributes_type
        )
        return attributes

    def remove_attributes(self, attributes_type: Type[ResourceAttributes]):
        """
        Remove the value of a given attributes type from this resource, if there is such a value.
        If the resource does not have a value for the given attributes type, do nothing.
        :param attributes_type:
        :return:
        """
        if not self._resource.has_attributes(attributes_type):
            return
        self._set_modified()
        self._resource.remove_attributes(attributes_type)

    def add_component(
        self,
        component_id: bytes,
        version: int,
    ):
        """
        Mark that a component has run on this resource

        :param component_id: ID of the component which ran
        :param version: Version of the component which ran
        :return:
        """
        self._set_modified()
        self._resource.add_component(component_id, version)

    def add_component_for_attributes(
        self,
        component_id: bytes,
        version: int,
        attributes: Type[ResourceAttributes],
    ):
        """
        Mark that a component was responsible for adding some attributes to this resource.
        :param component_id: ID of the component which added the attributes
        :param version: version of the component which added the attributes
        :param attributes: The type of attributes which were added
        :return:
        """
        self._set_modified()
        self._resource.add_component_for_attributes(component_id, version, attributes)

    def remove_component(
        self,
        component_id: bytes,
        attributes: Optional[Type[ResourceAttributes]] = None,
    ):
        """
        Remove any information that this component ran on this resource and/or added a particular
        type of attributes to this resource
        :param component_id: ID of the component to remove information about
        :param attributes: The type of attributes to remove information about
        :return:
        """
        self._set_modified()
        self._resource.remove_component(component_id, attributes)

    def has_component_run(self, component_id: bytes, desired_version: Optional[int] = None) -> bool:
        """
        Check if a particular component has run on this resource

        :param component_id: ID of the component to check for
        :param desired_version: If this is not `None`, also check that a specific version of
        ``component`` ran. Defaults to ``None``.
        :return: `True` if a component matching ``component_id`` and ``desired_version`` ran on
        this resource, `False` otherwise. If ``desired_version`` is `None`, only ``component_id``
        must be matched to return `True`.
        """
        version = self._resource.get_component_version(component_id)
        if version is None:
            return False
        if desired_version is None:
            return True
        return version == desired_version

    def queue_patch(
        self,
        patch_range: Range,
        data: bytes,
    ):
        """
        Replace the data within the provided range with the provided data. This operation may
        shrink, expand or leave untouched the resource's data. Patches are queued up to be
        applied, and will only be applied to the resource's data after the component this was
        called from exits.

        :param patch_range: The range of binary data in this resource to replace
        :param data: The bytes to replace part of this resource's data with
        :return:
        """
        if not self._component_context:
            raise InvalidStateError(
                f"Cannot patch resource {self._resource.id.hex()} without a context"
            )
        if self._resource.data_id is None:
            raise ValueError("Cannot patch a resource with no data")
        self._component_context.modification_trackers[self._resource.id].data_patches.append(
            DataPatch(
                patch_range,
                self._resource.data_id,
                data,
            )
        )
        self._resource.is_modified = True

    async def get_parent_as_view(self, v_type: Type[RV]) -> RV:
        """
        Get the parent of this resource. The parent will be returned as an instance of the given
        [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param v_type: The type of [view][ofrak.resource] to get the parent as
        """
        parent_r = await self.get_parent()
        return await parent_r.view_as(v_type)

    async def get_parent(self) -> "Resource":
        """
        Get the parent of this resource.
        """
        models = list(
            await self._resource_service.get_ancestors_by_id(self._resource.id, max_count=1)
        )
        if len(models) != 1:
            raise NotFoundError(f"There is no parent for resource {self._resource.id.hex()}")
        return await self._create_resource(models[0])

    async def get_ancestors(
        self,
        r_filter: ResourceFilter = None,
    ) -> Iterable["Resource"]:
        """
        Get all the ancestors of this resource. May optionally filter the ancestors so only those
        matching certain parameters are returned.

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter was provided and no resources match the provided filter
        """
        models = await self._resource_service.get_ancestors_by_id(
            self._resource.id, r_filter=r_filter
        )
        return await self._create_resources(models)

    async def get_only_ancestor_as_view(
        self,
        v_type: Type[RV],
        r_filter: ResourceFilter,
    ) -> RV:
        """
        Get the only ancestor of this resource which matches the given filter. The ancestor will be
        returned as an instance of the given
        [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If more or fewer than one ancestor matches ``r_filter``
        """
        ancestor_r = await self.get_only_ancestor(r_filter)
        return await ancestor_r.view_as(v_type)

    async def get_only_ancestor(self, r_filter: ResourceFilter) -> "Resource":
        """
        Get the only ancestor of this resource which matches the given filter.

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:
        """
        ancestors = list(
            await self._resource_service.get_ancestors_by_id(self._resource.id, 1, r_filter)
        )
        if len(ancestors) == 0:
            raise NotFoundError(
                f"There is no ancestor for resource {self._resource.id.hex()} matching the "
                f"provided filter"
            )
        return await self._create_resource(ancestors[0])

    async def get_descendants_as_view(
        self,
        v_type: Type[RV],
        max_depth: int = -1,
        r_filter: ResourceFilter = None,
        r_sort: ResourceSort = None,
    ) -> Iterable[RV]:
        """
        Get all the descendants of this resource. May optionally filter the descendants so only
        those matching certain parameters are returned. May optionally sort the descendants by
        an indexable attribute value key. The descendants will be returned as an
        instance of the given [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param v_type: The type of [view][ofrak.resource] to get the descendants as
        :param max_depth: Maximum depth from this resource to search for descendants; if -1,
        no maximum depth
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :param r_sort: Specifies which indexable attribute to use as the key to sort and the
        direction to sort
        :return:

        :raises NotFoundError: If a filter was provided and no resources match the provided filter
        """
        descendants = await self.get_descendants(max_depth, r_filter, r_sort)
        views_or_tasks = [r._view_as(v_type) for r in descendants]
        # analysis tasks to generate views of resources which don't have attrs for the view already
        view_tasks: List[Awaitable[RV]] = []
        # each resources' already-existing views OR the index in `view_tasks` of the analysis task
        views_or_task_indexes: List[Union[int, RV]] = []
        for view_or_create_view_task in views_or_tasks:
            if isawaitable(view_or_create_view_task):
                views_or_task_indexes.append(len(view_tasks))
                view_tasks.append(view_or_create_view_task)
            else:
                views_or_task_indexes.append(cast(RV, view_or_create_view_task))

        if view_tasks:
            completed_views: Sequence[RV] = await asyncio.gather(*view_tasks)
            return [
                completed_views[v_or_i] if type(v_or_i) is int else cast(RV, v_or_i)
                for v_or_i in views_or_task_indexes
            ]
        else:
            # There are no tasks, so all needed views are already present
            return cast(List[RV], views_or_task_indexes)

    async def get_descendants(
        self,
        max_depth: int = -1,
        r_filter: ResourceFilter = None,
        r_sort: ResourceSort = None,
    ) -> Iterable["Resource"]:
        """
        Get all the descendants of this resource. May optionally filter the descendants so only
        those matching certain parameters are returned. May optionally sort the descendants by
        an indexable attribute value key.

        :param max_depth: Maximum depth from this resource to search for descendants; if -1,
        no maximum depth
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :param r_sort: Specifies which indexable attribute to use as the key to sort and the
        direction to sort
        :return:

        :raises NotFoundError: If a filter was provided and no resources match the provided filter
        """
        models = await self._resource_service.get_descendants_by_id(
            self._resource.id, max_depth=max_depth, r_filter=r_filter, r_sort=r_sort
        )
        return await self._create_resources(models)

    async def get_only_descendant_as_view(
        self,
        v_type: Type[RV],
        max_depth: int = -1,
        r_filter: ResourceFilter = None,
    ) -> RV:
        """
        If a filter is provided, get the only descendant of this resource which matches the given
        filter. If a filter is not provided, gets the only descendant of this resource. The
        descendant will be returned as an instance of the given
        [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param v_type: The type of [view][ofrak.resource] to get the descendant as
        :param max_depth: Maximum depth from this resource to search for descendants; if -1,
        no maximum depth
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one descendant matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple descendant
        """
        descendant_r = await self.get_only_descendant(max_depth, r_filter)
        return await descendant_r.view_as(v_type)

    async def get_only_descendant(
        self,
        max_depth: int = -1,
        r_filter: ResourceFilter = None,
    ) -> "Resource":
        """
        If a filter is provided, get the only descendant of this resource which matches the given
        filter. If a filter is not provided, gets the only descendant of this resource.

        :param max_depth: Maximum depth from this resource to search for descendants; if -1,
        no maximum depth
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one descendant matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple descendant
        """
        models = list(
            await self._resource_service.get_descendants_by_id(
                self._resource.id,
                max_depth=max_depth,
                max_count=2,
                r_filter=r_filter,
            )
        )
        if len(models) == 0:
            raise NotFoundError(
                f"There is no descendant for resource {self._resource.id.hex()} matching "
                f"the provided filter {r_filter}"
            )
        if len(models) > 1:
            # TODO: Not the right kind of error
            raise NotFoundError(
                f"There are multiple descendants for resource {self._resource.id.hex()} "
                f"matching the provided filter"
            )
        return await self._create_resource(models[0])

    async def get_only_sibling_as_view(
        self,
        v_type: Type[RV],
        r_filter: ResourceFilter = None,
    ) -> RV:
        """
        If a filter is provided, get the only sibling of this resource which matches the given
        filter. If a filter is not provided, gets the only sibling of this resource. The sibling
        will be returned as an instance of the given
        [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].
        :param v_type: The type of [view][ofrak.resource] to get the sibling as
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one sibling matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple siblings
        """
        sibling_r = await self.get_only_sibling(r_filter)
        return await sibling_r.view_as(v_type)

    async def get_only_sibling(self, r_filter: ResourceFilter = None) -> "Resource":
        """
        If a filter is provided, get the only sibling of this resource which matches the given
        filter. If a filter is not provided, gets the only sibling of this resource.

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one sibling matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple siblings
        """
        models = list(
            await self._resource_service.get_siblings_by_id(
                self._resource.id,
                max_count=2,
                r_filter=r_filter,
            )
        )
        if len(models) == 0:
            raise NotFoundError(
                f"There is no sibling for resource {self._resource.id.hex()} matching "
                f"the provided filter"
            )
        if len(models) > 1:
            raise NotFoundError(
                f"There are multiple siblings for resource {self._resource.id.hex()} "
                f"matching the provided filter"
            )
        return await self._create_resource(models[0])

    async def get_children(
        self,
        r_filter: ResourceFilter = None,
        r_sort: ResourceSort = None,
    ) -> Iterable["Resource"]:
        """
        Get all the children of this resource. May optionally sort the children by an
        indexable attribute value key. May optionally filter the children so only those
        matching certain parameters are returned.

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :param r_sort: Specifies which indexable attribute to use as the key to sort and the
        direction to sort
        :return:

        :raises NotFoundError: If a filter was provided and no resources match the provided filter
        """
        return await self.get_descendants(1, r_filter, r_sort)

    async def get_children_as_view(
        self,
        v_type: Type[RV],
        r_filter: ResourceFilter = None,
        r_sort: ResourceSort = None,
    ) -> Iterable[RV]:
        """
        Get all the children of this resource. May optionally filter the children so only those
        matching certain parameters are returned. May optionally sort the children by an
        indexable attribute value key. The children will be returned as an instance of
        the given [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param v_type: The type of [view][ofrak.resource] to get the children as
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :param r_sort: Specifies which indexable attribute to use as the key to sort and the
        direction to sort
        :return:

        :raises NotFoundError: If a filter was provided and no resources match the provided filter
        """
        return await self.get_descendants_as_view(v_type, 1, r_filter, r_sort)

    async def get_only_child(self, r_filter: ResourceFilter = None) -> "Resource":
        """
        If a filter is provided, get the only child of this resource which matches the given
        filter. If a filter is not provided, gets the only child of this resource.

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one child matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple children
        """
        return await self.get_only_descendant(1, r_filter)

    async def get_only_child_as_view(self, v_type: Type[RV], r_filter: ResourceFilter = None) -> RV:
        """
        If a filter is provided, get the only child of this resource which matches the given
        filter. If a filter is not provided, gets the only child of this resource. The child will
        be returned as an instance of the given
        [viewable tag][ofrak.model.viewable_tag_model.ViewableResourceTag].

        :param v_type: The type of [view][ofrak.resource] to get the child as
        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :return:

        :raises NotFoundError: If a filter is provided and more or fewer than one child matches
        ``r_filter``
        :raises NotFoundError: If a filter is not provided and this resource has multiple children
        """
        return await self.get_only_descendant_as_view(v_type, 1, r_filter)

    async def delete(self):
        """
        Delete this resource and all of its descendants.

        :return:
        """
        self._component_context.resources_deleted.add(self._resource.id)

        for child_r in await self.get_children():
            await child_r.delete()

        self._resource.is_modified = True
        self._resource.is_deleted = True

    async def flush_to_disk(self, path: str, pack: bool = True):
        """
        Recursively repack the resource and write its data out to a file on disk. If this is a
        dataless resource, creates an empty file.

        :param path: Path to the file to write out to. The file is created if it does not exist.
        """
        if pack is True:
            await self.pack_recursively()

        data = await self.get_data()
        if data is not None:
            with open(path, "wb") as f:
                f.write(data)
        else:
            # Create empty file
            with open(path, "wb") as f:
                pass

    def __repr__(self):
        properties = [
            f"resource_id={self._resource.id.hex()}",
            f"tag=[{','.join([tag.__name__ for tag in self._resource.tags])}]",
        ]
        if self._resource.data_id:
            properties.append(f"data={self._resource.data_id.hex()}")
        return f"{type(self).__name__}(" + ", ".join(properties) + f")"

    async def summarize(self) -> str:
        """
        Create a string summary of this resource, including specific tags, attribute types,
        and the data offsets of this resource in the parent and root (if applicable).

        Not that this is not a complete string representation of the resource: not all tags are
        included, and only the types of attributes are included, not their values. It is a
        summary which gives a high level overview of the resource.
        """
        return await _default_summarize_resource(self)

    async def summarize_tree(
        self,
        r_filter: ResourceFilter = None,
        r_sort: ResourceSort = None,
        indent: str = "",
        summarize_resource_callback: Optional[Callable[["Resource"], Awaitable[str]]] = None,
    ) -> str:
        """
        Create a string summary of this resource and its (optionally filtered and/or sorted)
        descendants. The summaries of each resource are the same as the result of
        [summarize][ofrak.resource.Resource.summarize], organized into a tree structure.
        If a filter parameter is provided, it is applied recursively: the children of this
        resource will be filtered, then only those children matching
        the filter be displayed, and then the same filter will be applied to their children,
        etc. For example,

        :param r_filter: Contains parameters which resources must match to be returned, including
        any tags it must have and/or values of indexable attributes
        :param r_sort: Specifies which indexable attribute to use as the key to sort and the
        direction to sort
        """
        SPACER_BLANK = "   "
        SPACER_LINE = "───"

        if summarize_resource_callback is None:
            summarize_resource_callback = _default_summarize_resource

        children = cast(
            List[Resource], list(await self.get_children(r_filter=r_filter, r_sort=r_sort))
        )

        if children:
            if indent == "":
                tree_string = "┌"
            else:
                tree_string = "┬"
        else:
            tree_string = "─"

        tree_string += f"{await summarize_resource_callback(self)}\n"

        # All children but the last should display as a "fork" in the drop-down tree
        # After the last child, a vertical line should not be drawn as part of the indent
        # Both of those needs are handled here
        child_formatting: List[Tuple[str, str]] = [
            ("├", indent + "│" + SPACER_BLANK) for _ in children[:-1]
        ]
        child_formatting.append(("└", indent + " " + SPACER_BLANK))

        for child, (branch_symbol, child_indent) in zip(children, child_formatting):
            child_tree_string = await child.summarize_tree(
                r_filter=r_filter,
                r_sort=r_sort,
                indent=child_indent,
                summarize_resource_callback=summarize_resource_callback,
            )
            tree_string += f"{indent}{branch_symbol}{SPACER_LINE}{child_tree_string}"
        return tree_string


async def save_resources(
    resources: Iterable["Resource"],
    resource_service: ResourceServiceInterface,
    data_service: DataServiceInterface,
    component_context: ComponentContext,
    resource_context: ResourceContext,
    resource_view_context: ResourceViewContext,
):
    dependency_handler = DependencyHandler(
        resource_service, data_service, component_context, resource_context
    )
    resources_to_delete: List[bytes] = []
    patches_to_apply: List[DataPatch] = []
    resources_to_update: List[MutableResourceModel] = []

    for resource in resources:
        resource._save(
            resources_to_delete,
            patches_to_apply,
            resources_to_update,
        )

    deleted_descendants = await resource_service.delete_resources(resources_to_delete)
    data_ids_to_delete = [
        resource_m.data_id for resource_m in deleted_descendants if resource_m.data_id is not None
    ]
    await data_service.delete_models(data_ids_to_delete)
    patch_results = await data_service.apply_patches(patches_to_apply)
    await dependency_handler.handle_post_patch_dependencies(patch_results)
    diffs = []
    updated_ids = []
    for resource_m in resources_to_update:
        diffs.append(resource_m.save())
        updated_ids.append(resource_m.id)
    await resource_service.update_many(diffs)
    resource_view_context.update_views(updated_ids, resources_to_delete, resource_context)


class ResourceFactory:
    """
    Factory for creating [Resource][ofrak.resource.Resource].
    """

    def __init__(
        self,
        id_service: IDServiceInterface,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        job_service: JobServiceInterface,
    ):
        self._id_service = id_service
        self._data_service = data_service
        self._resource_service = resource_service
        self._job_service = job_service

    async def create(
        self,
        job_id: bytes,
        resource_id: bytes,
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        component_context: ComponentContext,
        job_context: Optional[JobRunContext] = None,
    ) -> Resource:
        """
        Create a resource from a resource_id.

        :param job_id:
        :param resource_id:
        :param resource_context:
        :param resource_view_context:
        :param component_context:
        :param job_context:
        """
        resource_m = resource_context.resource_models.get(resource_id)
        if resource_m is None:
            resource_m = MutableResourceModel.from_model(
                await self._resource_service.get_by_id(resource_id)
            )
            resource_context.resource_models[resource_id] = resource_m

        return next(
            iter(
                self._create(
                    job_id,
                    [resource_m],
                    resource_context,
                    resource_view_context,
                    component_context,
                    job_context,
                )
            )
        )

    async def create_many(
        self,
        job_id: bytes,
        resource_ids: Iterable[bytes],
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        component_context: ComponentContext,
        job_context: Optional[JobRunContext] = None,
    ) -> Iterable[Resource]:
        """
        Create Resources from resource_ids.

        :param job_id:
        :param resource_ids:
        :param resource_context:
        :param resource_view_context:
        :param component_context:
        :param job_context:
        """
        resource_models_minus_missing: List[Union[MutableResourceModel, int]] = []
        missing_ids: List[bytes] = []

        resource_m: MutableResourceModel

        for resource_id in resource_ids:
            resource_m = resource_context.resource_models.get(resource_id)  # type: ignore
            if resource_m is None:
                resource_models_minus_missing.append(len(missing_ids))
                missing_ids.append(resource_id)
            else:
                resource_models_minus_missing.append(resource_m)

        fetched_models: List[ResourceModel] = list(
            await self._resource_service.get_by_ids(missing_ids)
        )

        resource_models = []
        for resource_model_or_idx in resource_models_minus_missing:
            if type(resource_model_or_idx) is int:
                resource_m = MutableResourceModel.from_model(
                    fetched_models[cast(int, resource_model_or_idx)]
                )
                resource_models.append(resource_m)
                resource_context.resource_models[resource_m.id] = resource_m
            else:
                resource_models.append(cast(MutableResourceModel, resource_model_or_idx))

        return self._create(
            job_id,
            resource_models,
            resource_context,
            resource_view_context,
            component_context,
            job_context,
        )

    def _create(
        self,
        job_id: bytes,
        resource_models: List[MutableResourceModel],
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        component_context: ComponentContext,
        job_context: Optional[JobRunContext] = None,
    ) -> Iterable[Resource]:
        for resource_m in resource_models:
            yield Resource(
                job_id,
                resource_m,
                resource_context,
                resource_view_context,
                job_context,
                component_context,
                self,
                self._id_service,
                self._data_service,
                self._resource_service,
                self._job_service,
            )


async def _default_summarize_resource(resource: Resource) -> str:
    attributes_info = ", ".join(attrs_type.__name__ for attrs_type in resource._resource.attributes)

    if resource._resource.data_id:
        root_data_range = await resource.get_data_range_within_root()
        parent_data_range = await resource.get_data_range_within_parent()
        data = await resource.get_data()
        if len(data) <= 128:
            # Convert bytes to string to check .isprintable without doing .decode. Note that
            # not all ASCII is printable, so we have to check both decodable and printable
            raw_data_str = "".join(map(chr, data))
            if raw_data_str.isascii() and raw_data_str.isprintable():
                data_string = f'data_ascii="{data.decode("ascii")}"'
            else:
                data_string = f"data_hex={data.hex()}"
        else:
            sha256 = hashlib.sha256()
            sha256.update(data)
            data_string = f"data_hash={sha256.hexdigest()[:8]}"
        data_info = (
            f", global_offset=({hex(root_data_range.start)}-{hex(root_data_range.end)})"
            f", parent_offset=({hex(parent_data_range.start)}-{hex(parent_data_range.end)})"
            f", {data_string}"
        )
    else:
        data_info = ""
    return (
        f"{resource.get_id().hex()}: [caption=({resource.get_caption()}), "
        f"attributes=({attributes_info}){data_info}]"
    )



import logging
from abc import ABC, abstractmethod
from typing import Optional

from ofrak.component.abstract import AbstractComponent
from ofrak.model.component_model import CC
from ofrak.resource import Resource

LOGGER = logging.getLogger(__name__)


class IdentifierError(RuntimeError):
    """Base exception raised by identifiers."""


class Identifier(AbstractComponent[CC], ABC):
    """
    Identifiers are components that tag resources with specific resource tags.
    """

    @abstractmethod
    async def identify(self, resource: Resource, config: CC) -> None:
        """
        Perform identification on the given resource.

        Users should not call this method directly; rather, they should run
        [Resource.identify][ofrak.resource.Resource.identify].

        :param resource:
        :param config: Optional config for identifying. If an implementation provides a default,
        this default will always be used when config would otherwise be None. Note that a copy of
        the default config will be passed, so the default config values cannot be modified
        persistently by a component run.
        """
        raise NotImplementedError()

    @classmethod
    def get_default_config(cls) -> Optional[CC]:
        return cls._get_default_config_from_method(cls.identify)

    async def _run(self, resource: Resource, config: CC):
        if resource.has_component_run(self.get_id(), self.get_version()):
            LOGGER.debug(
                f"The {self.get_id().decode()} identifier has already been run on resource"
                f" {resource.get_id().hex()}"
            )
            return
        if resource.has_component_run(self.get_id()):
            # TODO: If the component has already ran (but on a different version), we should
            #  remove the tags it added
            LOGGER.warning(
                f"The {self.get_id().hex()} identifier has already been run on resource"
                f" {resource.get_id().hex()} under a different version. This is not "
                f"implemented yet."
            )
            raise NotImplementedError()
        await self.identify(resource, config)
        resource.add_component(self.get_id(), self.get_version())



import asyncio
import logging
import tempfile
from dataclasses import dataclass
from enum import Enum
from subprocess import CalledProcessError

from ofrak.component.analyzer import Analyzer
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier, Magic
from ofrak.model.component_model import ComponentExternalTool
from ofrak.resource import Resource
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

CPIO_TOOL = ComponentExternalTool(
    "cpio",
    "https://www.gnu.org/software/cpio/",
    install_check_arg="--help",
    apt_package="cpio",
    brew_package="cpio",
)


class CpioArchiveType(Enum):
    """
    CPIO has several unrelated, incompatible variants.
    They're described in the man page:
    https://linux.die.net/man/1/cpio
    """

    BINARY = "bin"
    OLD_ASCII = "odc"
    NEW_ASCII = "newc"
    CRC_ASCII = "crc"
    TAR = "tar"
    USTAR = "ustar"
    HPBIN = "hpbin"
    HPODC = "hpodc"


@dataclass
class CpioFilesystem(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a CPIO archive.
    """

    archive_type: CpioArchiveType


class CpioFilesystemAnalyzer(Analyzer[None, CpioFilesystem]):
    targets = (CpioFilesystem,)
    outputs = (CpioFilesystem,)

    async def analyze(self, resource: Resource, config=None) -> CpioFilesystem:
        _magic = await resource.analyze(Magic)
        magic_description = _magic.descriptor
        if magic_description.startswith("ASCII cpio archive (SVR4 with no CRC)"):
            archive_type = CpioArchiveType.NEW_ASCII
        elif magic_description.startswith("ASCII cpio archive (pre-SVR4 or odc)"):
            archive_type = CpioArchiveType.OLD_ASCII
        elif magic_description.startswith("ASCII cpio archive (SVR4 with CRC)"):
            archive_type = CpioArchiveType.CRC_ASCII
        elif magic_description.startswith("cpio archive"):
            archive_type = CpioArchiveType.BINARY
        else:
            raise NotImplementedError(
                f"Please add support for CPIO archive type {magic_description}"
            )

        return CpioFilesystem(archive_type)


class CpioUnpacker(Unpacker[None]):
    """
    Unpack a CPIO archive.
    """

    targets = (CpioFilesystem,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (CPIO_TOOL,)

    async def unpack(self, resource: Resource, config=None):
        cpio_v = await resource.view_as(CpioFilesystem)
        resource_data = await cpio_v.resource.get_data()
        with tempfile.TemporaryDirectory() as temp_flush_dir:
            cmd = [
                "cpio",
                "-id",
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                cwd=temp_flush_dir,
            )
            await proc.communicate(input=resource_data)
            if proc.returncode:
                raise CalledProcessError(returncode=proc.returncode, cmd=cmd)
            await cpio_v.initialize_from_disk(temp_flush_dir)


class CpioPacker(Packer[None]):
    """
    Pack files into a CPIO archive.
    """

    targets = (CpioFilesystem,)
    external_dependencies = (CPIO_TOOL,)

    async def pack(self, resource: Resource, config=None):
        cpio_v: CpioFilesystem = await resource.view_as(CpioFilesystem)
        temp_flush_dir = await cpio_v.flush_to_disk()
        cpio_format = cpio_v.archive_type.value
        list_files_cmd = [
            "find",
            "-print",
        ]
        list_files_proc = await asyncio.create_subprocess_exec(
            *list_files_cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=temp_flush_dir,
        )
        list_files_list, stderr = await list_files_proc.communicate()
        if list_files_proc.returncode:
            raise CalledProcessError(returncode=list_files_proc.returncode, cmd=list_files_cmd)

        cpio_pack_cmd = [
            "cpio",
            "-o",
            f"--format={cpio_format}",
        ]
        cpio_pack_proc = await asyncio.create_subprocess_exec(
            *cpio_pack_cmd,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=temp_flush_dir,
        )
        cpio_pack_output, stderr = await cpio_pack_proc.communicate(input=list_files_list)
        if cpio_pack_proc.returncode:
            raise CalledProcessError(returncode=cpio_pack_proc.returncode, cmd=cpio_pack_cmd)
        # Passing in the original range effectively replaces the original data with the new data
        resource.queue_patch(Range(0, await resource.get_data_length()), cpio_pack_output)


MagicMimeIdentifier.register(CpioFilesystem, "application/x-cpio")
MagicDescriptionIdentifier.register(CpioFilesystem, lambda s: "cpio archive" in s)



import os
import subprocess

import pytest

MAKEFILE_CONTENTS = """
CC=gcc

default: program

program.o: program.c $(HEADERS)
	$(CC) -c program.c -fno-asynchronous-unwind-tables -o program.o

program: program.o
	$(CC) program.o -o program

program_no_reloc: program.o
	$(CC) program.o -no-pie -o program_no_reloc

program_relocated: program_relocated.o
	$(CC) program_relocated.o -o program_relocated
"""

C_SOURCE_CONTENTS = """
#include <stdio.h>

int foo();
int bar();

int main() {
   printf("Hello, World!\\n");
   return foo();
}

int foo() {
    return 12;
}

int bar() {
    return 24;
}


"""


@pytest.fixture
def elf_test_directory(tmpdir):
    makefile_path = os.path.join(tmpdir, "Makefile")
    c_source_path = os.path.join(tmpdir, "program.c")

    with open(makefile_path, "w") as f:
        f.write(MAKEFILE_CONTENTS)
    with open(c_source_path, "w") as f:
        f.write(C_SOURCE_CONTENTS)

    return tmpdir


@pytest.fixture
def elf_object_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program.o"])
    return os.path.join(elf_test_directory, "program.o")


@pytest.fixture
def elf_executable_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program"])
    return os.path.join(elf_test_directory, "program")


@pytest.fixture
def elf_no_pie_executable_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program_no_reloc"])
    return os.path.join(elf_test_directory, "program_no_reloc")



import functools
import inspect


def auto_validate_state(validator_function):
    """
    Class decorator for services or other objects with complex internal state. Used to create
    classes with methods that run a state validation function after every (public) method call.

    Public methods are detected as any callable whose name does not start with an underscore ("_").

    Example usage:
    def validator_function(class_instance):
        # validation logic, raising an error if the internal state is invalid
        ...


    @auto_validate_state(validator_function)
    class AutoValidatingImplementation(ClassWithInternalState):
        pass

    An instance of AutoValidatingImplementation now runs validator_function on itself after every
    method call.

    :param validator_function: Function to check internal state of the decorated class, and should
    raise a descriptive error if the state is invalid

    :return:
    """

    def transform_method_to_validate_state(method):
        if inspect.iscoroutinefunction(method):

            @functools.wraps(method)
            async def wrapped_method(self, *args, **kwargs):
                result = await method(self, *args, **kwargs)
                validator_function(self)

                return result

        else:

            @functools.wraps(method)
            def wrapped_method(self, *args, **kwargs):
                result = method(self, *args, **kwargs)
                validator_function(self)

                return result

        return wrapped_method

    def decorator(cls):
        methods = [
            (attr_name, getattr(cls, attr_name))
            for attr_name in dir(cls)
            if not attr_name.startswith("_") and callable(getattr(cls, attr_name))
        ]
        for method_name, method in methods:
            setattr(cls, method_name, transform_method_to_validate_state(method))

        return cls

    return decorator



from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import List, Tuple, Dict, Optional, Iterable, Mapping

from immutabledict import immutabledict

from ofrak.core.binary import BinaryPatchModifier, BinaryPatchConfig

from ofrak.component.analyzer import Analyzer
from ofrak.component.modifier import Modifier, ModifierError
from ofrak.core.memory_region import MemoryRegion
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import index
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeRangeFilter,
    ResourceAttributeValueFilter,
    ResourceSort,
)
from ofrak_patch_maker.model import AssembledObject, PatchRegionConfig, BOM
from ofrak_patch_maker.toolchain.model import Segment
from ofrak_type.memory_permissions import MemoryPermissions
from ofrak_type.range import Range, remove_subranges


class FreeSpaceAllocationError(RuntimeError):
    pass


@dataclass
class FreeSpace(MemoryRegion):
    permissions: MemoryPermissions

    @index
    def Permissions(self) -> int:
        return self.permissions.value


@dataclass
class FreeSpaceAllocation(ComponentConfig):
    permissions: MemoryPermissions
    allocations: List[Range]


@dataclass
class Allocatable(ResourceView):
    """
    Identifies a resource that may have some free space available within it to allocate (for
    arbitrary purposes). Once tagged as an Allocatable, a resource may be analyzed to find all of
    the available free space within it and its descendants. These ranges are stored as
    minimal lists of contiguous ranges of free space (ranges which overlap or touch are combined),
    sorted first by size then vaddr (lowest to highest). Each list of free space ranges contains
    only free space with one type of memory access permissions of the ranges, i.e. all read-only
    free space is stored in one list, all read-execute free-space is stored in another list, etc.
    """

    free_space_ranges: Dict[MemoryPermissions, List[Range]]

    async def allocate(
        self,
        permissions: MemoryPermissions,
        requested_size: int,
        alignment: int = 4,
        min_fragment_size: Optional[int] = None,
        within_range: Optional[Range] = None,
    ) -> List[Range]:
        """
        Request some range(s) of free space which satisfies the given constraints as parameters.
        If such ranges are found, the resources they belong to (which are descendants of this
        `Allocatable`) are modified to reflect that part or all of them are no longer valid free
        space using
        [RemoveFreeSpaceModifier][ofrak.core.free_space.RemoveFreeSpaceModifier].

        Allows either a fragmented allocation (X space is allocated across N individual ranges),
        or a non-fragmented allocation (one range of size X). To force the allocation to be
        non-fragmented, set the `min_fragment_size` param equal to the `requested_size`.

        :param permissions: Required memory permissions of the free space (exact match)
        :param requested_size: Total size of allocated ranges
        :param alignment: That start of the allocated ranges will be aligned to `alignment` bytes
        :param min_fragment_size: The minimum size of each allocated range
        :param within_range: All returned ranges must be within this virtual address range

        :return: A list of one or more [Range][ofrak_type.range.Range], which each contain a
         start and end vaddr of an allocated range of free space

        :raises FreeSpaceAllocationError: if there is not enough free space to allocate which
        matches the given constraints
        """
        allocated_ranges = await self._allocate(
            permissions,
            requested_size,
            alignment,
            min_fragment_size,
            within_range,
        )

        # Having acquired a satisfactory allocation, make sure subsequent calls won't allocate
        # from this same block
        await self.resource.run(
            RemoveFreeSpaceModifier,
            FreeSpaceAllocation(
                permissions,
                allocated_ranges,
            ),
        )
        self.remove_allocation_from_cached_free_ranges(allocated_ranges, permissions)

        return allocated_ranges

    async def allocate_bom(
        self,
        bom: BOM,
        permission_map: Optional[Mapping[MemoryPermissions, Iterable[MemoryPermissions]]] = None,
    ) -> PatchRegionConfig:
        """
        Responsible for allocating the patches if free memory is required and
        providing details about where space was made.

        :param bom:
        :param permission_map: a map that assigns patch segment-local permissions to possible
        destination permissions. available memory pool is evaluated in order!
        Ex: a developer wants to enable allocation of RO strings from .rodata in an RX .text section.

        :return: information required to generate the linker directive script
        """
        segments_to_allocate: List[Tuple[AssembledObject, Segment]] = []
        for obj in bom.object_map.values():
            for segment in obj.segment_map.values():
                segments_to_allocate.append((obj, segment))

        # Allocate largest segments first
        segments_to_allocate.sort(key=lambda o_s: o_s[1].length, reverse=True)
        segments_by_object: Dict[str, List[Segment]] = defaultdict(list)
        for obj, segment in segments_to_allocate:
            vaddr, final_size = 0, 0
            if segment.length == 0:
                continue
            if permission_map is not None:
                possible_perms = permission_map[segment.access_perms]
            else:
                possible_perms = (segment.access_perms,)
            for candidate_permissions in possible_perms:
                try:
                    allocs = await self.allocate(
                        candidate_permissions,
                        segment.length,
                        min_fragment_size=segment.length,
                        alignment=bom.segment_alignment,
                    )
                    allocation = next(iter(allocs))
                    vaddr = allocation.start
                    final_size = allocation.length()
                    break
                except FreeSpaceAllocationError:
                    continue
            if vaddr == 0 or final_size == 0:
                raise FreeSpaceAllocationError(
                    f"Could not find enough free space for access perms {possible_perms} and "
                    f"length {segment.length}"
                )
            segments_by_object[obj.path].append(
                Segment(
                    segment_name=segment.segment_name,
                    vm_address=vaddr,
                    offset=segment.offset,
                    is_entry=segment.is_entry,
                    length=final_size,
                    access_perms=segment.access_perms,
                )
            )

        all_segments: Dict[str, Tuple[Segment, ...]] = {
            object_path: tuple(segments) for object_path, segments in segments_by_object.items()
        }

        return PatchRegionConfig(bom.name + "_patch", immutabledict(all_segments))

    async def _allocate(
        self,
        permissions: MemoryPermissions,
        requested_size: int,
        alignment: int = 4,
        min_fragment_size: Optional[int] = None,
        within_range: Optional[Range] = None,
    ) -> List[Range]:
        free_ranges = self.free_space_ranges.get(permissions)
        if not free_ranges:
            raise FreeSpaceAllocationError(f"No free space with permissions {permissions}.")

        if min_fragment_size is None:
            _min_fragment_size = requested_size
        elif min_fragment_size <= 0:
            raise ValueError(f"Minimum fragment size must be >0 or None! Got {min_fragment_size}.")
        else:
            _min_fragment_size = min_fragment_size

        unallocated_size = requested_size
        allocation: List = list()

        for free_range in free_ranges:
            assert unallocated_size >= 0
            if unallocated_size == 0:
                break

            # 1. Transform free range to only ranges within vaddr range constraint
            try:
                if within_range:
                    free_range = free_range.intersect(within_range)
            except ValueError:
                # There is no overlap with the `within_range` constraint
                continue
            # 2. Transform free range to satisfy alignment
            try:
                free_range = self._align_range(free_range, alignment)
            except ValueError:
                # Free range is too small to satisfy alignment
                continue
            # 3. Transform free range to only the range we actually require for allocation
            new_length = min(free_range.length(), unallocated_size)
            free_range = Range.from_size(free_range.start, new_length)
            # If transformed range is not not big enough, continue
            if free_range.length() < _min_fragment_size:
                continue

            allocation.append(free_range)
            unallocated_size -= free_range.length()

        assert unallocated_size >= 0
        if unallocated_size > 0:
            raise FreeSpaceAllocationError(
                f"Not enough valid free space to allocate {requested_size} "
                f"units with constraints alignment={alignment}, "
                f"permissions={permissions}, "
                f"min_fragment_size={min_fragment_size}, "
                f"within_range={within_range}."
            )

        return allocation

    @staticmethod
    def _align_range(unaligned_range: Range, alignment: int) -> Range:
        offset_to_align_start = (alignment - (unaligned_range.start % alignment)) % alignment
        # Currently we don't expect the end of each allocated range to be aligned
        # If we end up wanting to align both start and end, `offset_to_align_end` should be updated
        offset_to_align_end = 0
        aligned_range = Range(
            unaligned_range.start + offset_to_align_start,
            unaligned_range.end + offset_to_align_end,
        )
        assert aligned_range.start >= unaligned_range.start
        assert aligned_range.end <= unaligned_range.end

        return aligned_range

    @staticmethod
    def sort_free_ranges(ranges: Iterable[Range]) -> List[Range]:
        return list(sorted(ranges, key=lambda r: (r.length(), r.start)))

    def remove_allocation_from_cached_free_ranges(
        self, allocation: List[Range], permissions: MemoryPermissions
    ):
        if len(allocation) == 0:
            return
        new_free_ranges = remove_subranges(
            self.free_space_ranges[permissions],
            allocation,
        )

        self.free_space_ranges[permissions] = Allocatable.sort_free_ranges(new_free_ranges)


class FreeSpaceAnalyzer(Analyzer[None, Allocatable]):
    """
    Analyze an `Allocatable` resource to find the ranges of free space it contains by searching for
    descendants tagged as `FreeSpace`. The ranges of each individual `FreeSpace` resource will be
    globbed into as few non-overlapping ranges as possible. The ranges of different types of free
    space - such as RW permissions vs RX permissions - will be calculated and stored separately.
    """

    targets = (Allocatable,)
    outputs = (Allocatable,)

    async def analyze(self, resource: Resource, config: None) -> Allocatable:
        ranges_by_permissions = defaultdict(list)
        for free_space_r in await resource.get_descendants_as_view(
            FreeSpace,
            r_filter=ResourceFilter.with_tags(FreeSpace),
            r_sort=ResourceSort(FreeSpace.VirtualAddress),
        ):
            ranges_by_permissions[free_space_r.permissions].append(free_space_r.vaddr_range())

        merged_ranges_by_permissions = dict()
        for perms, ranges in ranges_by_permissions.items():
            merged_ranges_by_permissions[perms] = Allocatable.sort_free_ranges(
                Range.merge_ranges(ranges)
            )

        return Allocatable(merged_ranges_by_permissions)


class RemoveFreeSpaceModifier(Modifier[FreeSpaceAllocation]):
    """
    After allocating some space from an `Allocatable`, fix up its descendants to make sure the
    allocated space will not be allocated again. Remove FreeSpace tags from resources which
    overlap with an allocated range. If part of one of these resources is not within an
    allocated range, create a child tagged as FreeSpace to reflect that part of it is still
    available as free space.
    """

    targets = (Allocatable,)

    async def modify(self, resource: Resource, config: FreeSpaceAllocation) -> None:

        wholly_allocated_resources = list()
        partially_allocated_resources: Dict[bytes, Tuple[FreeSpace, List[Range]]] = dict()
        allocatable = await resource.view_as(Allocatable)

        for alloc in config.allocations:
            for res_wholly_in_alloc in await resource.get_descendants_as_view(
                FreeSpace,
                r_filter=ResourceFilter(
                    tags=(FreeSpace,),
                    attribute_filters=(
                        ResourceAttributeValueFilter(
                            FreeSpace.Permissions, config.permissions.value
                        ),
                        ResourceAttributeRangeFilter(
                            FreeSpace.VirtualAddress,
                            min=alloc.start,
                            max=alloc.end - 1,
                        ),
                        ResourceAttributeRangeFilter(
                            FreeSpace.EndVaddr, min=alloc.start + 1, max=alloc.end
                        ),
                    ),
                ),
            ):
                wholly_allocated_resources.append(res_wholly_in_alloc)

            for res_partially_in_alloc in await self._get_partially_overlapping_resources(
                resource,
                config.permissions,
                alloc,
            ):
                free_space_range = res_partially_in_alloc.vaddr_range()
                overlap = alloc.intersect(free_space_range)
                assert overlap.length() > 0
                free_space_res_id = res_partially_in_alloc.resource.get_id()
                if free_space_res_id in partially_allocated_resources:
                    _, allocated_ranges_of_res = partially_allocated_resources[free_space_res_id]
                    allocated_ranges_of_res.append(overlap)
                else:
                    partially_allocated_resources[free_space_res_id] = (
                        res_partially_in_alloc,
                        [overlap],
                    )

        for fs in wholly_allocated_resources:
            fs.resource.remove_tag(FreeSpace)

        for fs, allocated_ranges in partially_allocated_resources.values():
            remaining_free_space_ranges = remove_subranges([fs.vaddr_range()], allocated_ranges)
            for remaining_range in remaining_free_space_ranges:
                remaining_data_range = Range.from_size(
                    fs.get_offset_in_self(remaining_range.start), remaining_range.length()
                )
                await fs.resource.create_child_from_view(
                    FreeSpace(
                        remaining_range.start,
                        remaining_range.length(),
                        fs.permissions,
                    ),
                    data_range=remaining_data_range,
                )
            fs.resource.remove_tag(FreeSpace)

        # Update Allocatable attributes, reflecting removed ranges
        allocatable.remove_allocation_from_cached_free_ranges(
            config.allocations, config.permissions
        )
        resource.add_view(allocatable)

    @staticmethod
    async def _get_partially_overlapping_resources(
        resource: Resource,
        permissions: MemoryPermissions,
        alloc: Range,
    ) -> Iterable[FreeSpace]:
        filter_overlapping_free_range_end = (
            ResourceAttributeValueFilter(FreeSpace.Permissions, permissions.value),
            ResourceAttributeRangeFilter(FreeSpace.VirtualAddress, max=alloc.end),
            ResourceAttributeRangeFilter(
                FreeSpace.EndVaddr,
                min=alloc.end + 1,
            ),
        )
        filter_overlapping_free_range_start = (
            ResourceAttributeValueFilter(FreeSpace.Permissions, permissions.value),
            ResourceAttributeRangeFilter(
                FreeSpace.VirtualAddress,
                max=alloc.start - 1,
            ),
            ResourceAttributeRangeFilter(
                FreeSpace.EndVaddr,
                min=alloc.start + 1,
            ),
        )

        resources_overlapping_free_range_end = await resource.get_descendants_as_view(
            FreeSpace,
            r_filter=ResourceFilter(
                tags=(FreeSpace,),
                attribute_filters=filter_overlapping_free_range_end,
            ),
        )
        resources_overlapping_free_range_start = await resource.get_descendants_as_view(
            FreeSpace,
            r_filter=ResourceFilter(
                tags=(FreeSpace,),
                attribute_filters=filter_overlapping_free_range_start,
            ),
        )

        return chain(
            resources_overlapping_free_range_end,
            resources_overlapping_free_range_start,
        )


async def _find_and_delete_overlapping_children(resource: Resource, freed_range: Range):
    # Note this filter calculation has the potential to be very expensive if, for instance,
    # the resource is an entire program segment...
    overlap_resources = list(
        await resource.get_children_as_view(
            MemoryRegion,
            r_filter=ResourceFilter(
                tags=(MemoryRegion,),
                attribute_filters=(
                    ResourceAttributeRangeFilter(MemoryRegion.VirtualAddress, max=freed_range.end),
                    ResourceAttributeRangeFilter(MemoryRegion.EndVaddr, min=freed_range.start),
                ),
            ),
        )
    )
    for overlapping_child in overlap_resources:
        await overlapping_child.resource.delete()
        await overlapping_child.resource.save()


def _get_fill(freed_range: Range, fill: Optional[bytes]):
    if not fill:
        return b"\x00" * freed_range.length()
    else:
        diff_len = freed_range.length() - len(fill)
        if diff_len < 0:
            raise ValueError("config.fill value cannot be longer than the range to be freed.")
        return fill + b"\x00" * diff_len


@dataclass
class FreeSpaceModifierConfig(ComponentConfig):
    """
    Configuration for modifier which marks some free space.

    :var permissions: memory permissions to give the created free space.
    :var fill: bytes to fill the free space with
    """

    permissions: MemoryPermissions
    fill: Optional[bytes] = None


class FreeSpaceModifier(Modifier[FreeSpaceModifierConfig]):
    """
    Turn a [MemoryRegion][ofrak.core.memory_region.MemoryRegion] resource into allocatable free
    space by replacing its data with b'\x00' or optionally specified bytes.
    [FreeSpace][ofrak.core.free_space.FreeSpace].
    """

    targets = (MemoryRegion,)

    async def modify(self, resource: Resource, config: FreeSpaceModifierConfig):
        mem_region_view = await resource.view_as(MemoryRegion)

        freed_range = Range(
            mem_region_view.virtual_address,
            mem_region_view.virtual_address + mem_region_view.size,
        )
        patch_data = _get_fill(freed_range, config.fill)
        parent_mr_view = await resource.get_parent_as_view(MemoryRegion)
        patch_offset = parent_mr_view.get_offset_in_self(freed_range.start)
        patch_range = freed_range.translate(patch_offset - freed_range.start)

        await resource.delete()
        await resource.save()

        # Patch in the patch_data
        await parent_mr_view.resource.run(
            BinaryPatchModifier, BinaryPatchConfig(patch_offset, patch_data)
        )
        # Create the FreeSpace child
        await parent_mr_view.resource.create_child_from_view(
            FreeSpace(
                mem_region_view.virtual_address,
                mem_region_view.size,
                config.permissions,
            ),
            data_range=patch_range,
        )


@dataclass
class PartialFreeSpaceModifierConfig(ComponentConfig):
    """
    :var permissions: memory permissions to give the created free space.
    :var range_to_remove: the ranges to consider as free space (remove)
    :var fill: bytes to fill the free space with
    """

    permissions: MemoryPermissions
    range_to_remove: Range
    fill: Optional[bytes] = None


class PartialFreeSpaceModifier(Modifier[PartialFreeSpaceModifierConfig]):
    """
    Turn part of a [MemoryRegion][ofrak.core.memory_region.MemoryRegion] resource into allocatable
    free space by replacing a range of its data with b'\x00' or optionally specified fill bytes.
    [FreeSpace][ofrak.core.free_space.FreeSpace] child resource at that range.
    """

    targets = (MemoryRegion,)

    async def modify(self, resource: Resource, config: PartialFreeSpaceModifierConfig):
        freed_range = config.range_to_remove
        mem_region_view = await resource.view_as(MemoryRegion)
        if not freed_range.within(mem_region_view.vaddr_range()):
            raise ModifierError(
                f"Free space range, {freed_range}, must lie within target memory"
                f"region range, {mem_region_view.vaddr_range()}"
            )

        await _find_and_delete_overlapping_children(resource, freed_range)

        patch_offset = mem_region_view.get_offset_in_self(freed_range.start)
        patch_range = Range.from_size(patch_offset, freed_range.length())
        patch_data = _get_fill(freed_range, config.fill)
        await mem_region_view.resource.run(
            BinaryPatchModifier, BinaryPatchConfig(patch_offset, patch_data)
        )
        await mem_region_view.resource.create_child_from_view(
            FreeSpace(
                freed_range.start,
                freed_range.length(),
                config.permissions,
            ),
            data_range=patch_range,
        )



from dataclasses import dataclass

from ofrak.core.memory_region import MemoryRegion
from ofrak.model.resource_model import index
from ofrak.model.viewable_tag_model import AttributesType


@dataclass
class ProgramSection(MemoryRegion):
    """
    A section of a program.

    :ivar virtual_address: The virtual address at the start of the section
    :ivar size: The size of the section in bytes
    """


@dataclass
class NamedProgramSection(ProgramSection):
    """
    A section of a program with a name.

    :ivar virtual_address: The virtual address at the start of the section
    :ivar size: The size of the section in bytes
    :ivar name: The name of the section
    """

    name: str

    @index
    def SectionName(self) -> str:
        return self.name

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            namedprogram_attributes = all_attributes[AttributesType[NamedProgramSection]]
        except KeyError:
            return super().caption(all_attributes)
        return f"{str(cls.__name__)}: {namedprogram_attributes.name}"


@dataclass
class ProgramSegment(MemoryRegion):
    """
    A segment of a program.

    :ivar virtual_address: The virtual address at the start of the segment
    :ivar size: The size of the segment in bytes
    """



from typing import Any, Dict, Tuple, List, Union

from typing_inspect import get_origin, get_args

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class DictSerializer(SerializerInterface):
    """
    Serialize and deserialize `Dict[X, Y]` into `PJSONType`.

    Implementation: a list of tuples (key, value), because the serialized keys may not be hashable
    (e.g. if they're themselves serialized as a dictionary), despite the real keys being hashable.
    """

    targets = (lambda type_hint: get_origin(type_hint) == dict,)

    # Tuples are kept as tuples in PJSON, but serialized into lists in JSON.
    SerializedDictType = Union[List[Tuple[PJSONType, PJSONType]], List[List[PJSONType]]]

    def obj_to_pjson(self, obj: Dict[Any, Any], type_hint: Any) -> SerializedDictType:
        key_type, value_type = get_args(type_hint)
        return [
            (self._service.to_pjson(key, key_type), self._service.to_pjson(value, value_type))
            for key, value in obj.items()
        ]

    def pjson_to_obj(self, pjson_obj: SerializedDictType, type_hint: Any) -> Dict[Any, Any]:
        key_type, value_type = get_args(type_hint)
        return {
            self._service.from_pjson(key, key_type): self._service.from_pjson(value, value_type)
            for key, value in pjson_obj
        }



import asyncio
import logging
import os
import tempfile
import time
from types import ModuleType
from typing import Type, Any, Awaitable, Callable, List, Iterable, Optional

from ofrak_type import InvalidStateError
from synthol.injector import DependencyInjector

from ofrak.component.interface import ComponentInterface
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File, FilesystemRoot
from ofrak.model.component_model import ClientComponentContext
from ofrak.model.resource_model import ResourceModel, ClientResourceContextFactory
from ofrak.model.tag_model import ResourceTag
from ofrak.model.viewable_tag_model import ResourceViewContext
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.abstract_ofrak_service import AbstractOfrakService
from ofrak.service.component_locator_i import ComponentLocatorInterface
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.id_service_i import IDServiceInterface
from ofrak.service.job_service_i import JobServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface

LOGGER = logging.getLogger("ofrak")
DEFAULT_OFRAK_LOG_FILE = os.path.join(tempfile.gettempdir(), "ofrak.log")


class OFRAKContext:
    def __init__(
        self,
        injector: DependencyInjector,
        resource_factory: ResourceFactory,
        component_locator: ComponentLocatorInterface,
        id_service: IDServiceInterface,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        job_service: JobServiceInterface,
        all_ofrak_services: List[AbstractOfrakService],
    ):
        self.injector = injector
        self.resource_factory = resource_factory
        self.component_locator = component_locator
        self.id_service = id_service
        self.data_service = data_service
        self.resource_service = resource_service
        self.job_service = job_service
        self._all_ofrak_services = all_ofrak_services
        self._resource_context_factory = ClientResourceContextFactory()

    async def create_root_resource(
        self, name: str, data: bytes, tags: Iterable[ResourceTag] = (GenericBinary,)
    ) -> Resource:
        job_id = self.id_service.generate_id()
        resource_id = self.id_service.generate_id()
        data_id = resource_id

        await self.job_service.create_job(job_id, name)
        await self.data_service.create_root(data_id, data)
        resource_model = await self.resource_service.create(
            ResourceModel.create(resource_id, data_id, tags=tags)
        )
        root_resource = await self.resource_factory.create(
            job_id,
            resource_model.id,
            self._resource_context_factory.create(),
            ResourceViewContext(),
            ClientComponentContext(),
        )
        return root_resource

    async def create_root_resource_from_file(self, file_path: str) -> Resource:
        full_file_path = os.path.abspath(file_path)
        with open(full_file_path, "rb") as f:
            root_resource = await self.create_root_resource(
                os.path.basename(full_file_path), f.read(), (File,)
            )
        root_resource.add_view(
            File(
                os.path.basename(full_file_path),
                os.lstat(full_file_path),
                FilesystemRoot._get_xattr_map(full_file_path),
            )
        )
        await root_resource.save()
        return root_resource

    async def start_context(self):
        if "_ofrak_context" in globals():
            raise InvalidStateError(
                "Cannot start OFRAK context as a context has already been started in this process!"
            )
        globals()["_ofrak_context"] = self
        await asyncio.gather(*(service.run() for service in self._all_ofrak_services))

    async def shutdown_context(self):
        if "_ofrak_context" in globals():
            del globals()["_ofrak_context"]
        await asyncio.gather(*(service.shutdown() for service in self._all_ofrak_services))
        logging.shutdown()

    def get_all_tags(self) -> Iterable[ResourceTag]:
        all_tags = ResourceTag.all_tags
        return all_tags


class OFRAK:
    DEFAULT_LOG_LEVEL = logging.WARNING

    def __init__(
        self,
        logging_level: int = DEFAULT_LOG_LEVEL,
        exclude_components_missing_dependencies: bool = False,
    ):
        """
        Set up the OFRAK environment that a script will use.

        :param logging_level: Logging level of OFRAK instance (logging.DEBUG, logging.WARNING, etc.)
        :param exclude_components_missing_dependencies: When initializing OFRAK, check each component's dependency and do
        not use any components missing some dependencies
        """
        logging.basicConfig(level=logging_level, format="[%(filename)15s:%(lineno)5s] %(message)s")
        logging.getLogger().addHandler(logging.FileHandler(DEFAULT_OFRAK_LOG_FILE))
        logging.getLogger().setLevel(logging_level)
        logging.captureWarnings(True)
        self.injector = DependencyInjector()
        self._discovered_modules: List[ModuleType] = []
        self._exclude_components_missing_dependencies = exclude_components_missing_dependencies
        self._id_service: Optional[IDServiceInterface] = None

    def discover(
        self,
        module: ModuleType,
        blacklisted_interfaces: Iterable[Type] = (),
        blacklisted_modules: Iterable[Any] = (),
    ):
        self.injector.discover(module, blacklisted_interfaces, blacklisted_modules)
        self._discovered_modules.append(module)

    def set_id_service(self, service: IDServiceInterface):
        self._id_service = service

    async def create_ofrak_context(self) -> OFRAKContext:
        """
        Create the OFRAKContext and start all its services.
        """
        self._setup()
        component_locator = await self.injector.get_instance(ComponentLocatorInterface)

        resource_factory = await self.injector.get_instance(ResourceFactory)
        components = await self._get_discovered_components()
        component_locator.add_components(components, self._discovered_modules)

        id_service = await self.injector.get_instance(IDServiceInterface)
        data_service = await self.injector.get_instance(DataServiceInterface)
        resource_service = await self.injector.get_instance(ResourceServiceInterface)
        job_service = await self.injector.get_instance(JobServiceInterface)
        all_services = await self.injector.get_instance(List[AbstractOfrakService])

        ofrak_context = OFRAKContext(
            self.injector,
            resource_factory,
            component_locator,
            id_service,
            data_service,
            resource_service,
            job_service,
            all_services,
        )
        await ofrak_context.start_context()
        return ofrak_context

    # TODO: Typehints here do not properly accept functions with variable args
    async def run_async(self, func: Callable[["OFRAKContext", Any], Awaitable[None]], *args):
        ofrak_context = await self.create_ofrak_context()
        start = time.time()
        try:
            await func(ofrak_context, *args)
        finally:
            await ofrak_context.shutdown_context()
            print(f"It took {time.time() - start:.3f} seconds to run the OFRAK script")

    # TODO: Typehints here do not properly accept functions with variable args
    def run(self, func: Callable[["OFRAKContext", Any], Awaitable[None]], *args):
        asyncio.get_event_loop().run_until_complete(self.run_async(func, *args))

    def _setup(self):
        """Discover common OFRAK services and components."""
        import ofrak

        self.discover(ofrak)

        if self._id_service:
            self.injector.bind_instance(self._id_service)

    async def _get_discovered_components(self) -> List[ComponentInterface]:
        all_discovered_components = await self.injector.get_instance(List[ComponentInterface])
        if not self._exclude_components_missing_dependencies:
            return all_discovered_components
        LOGGER.debug(
            "`exclude_components_missing_dependencies` set True; checking each discovered component's dependencies are "
            "installed"
        )
        components_missing_deps = []
        audited_components = []
        for component in all_discovered_components:
            if all(
                await asyncio.gather(
                    *[dep.is_tool_installed() for dep in component.external_dependencies]
                )
            ):
                audited_components.append(component)
            else:
                components_missing_deps.append(component)

        LOGGER.warning(
            f"Skipped registering the following components due to missing dependencies: "
            f"{', '.join(type(c).__name__ for c in components_missing_deps)}. Run `python3 -m "
            f"ofrak deps --missing-only` for more details."
        )

        return audited_components


def get_current_ofrak_context() -> OFRAKContext:
    # TODO: This is a brittle MVP, creating multiple simultaneous contexts in a single process
    #  will probably break it!
    ctx = globals().get("_ofrak_context")
    if ctx is None:
        raise InvalidStateError("Not in an OFRAK context!")
    else:
        return ctx



from dataclasses import dataclass
from typing import Type, Tuple, Optional

from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.tag_model import ResourceTag


@dataclass
class JobAnalyzerRequest:
    job_id: bytes
    resource_id: bytes
    attributes: Type[ResourceAttributes]
    target_tags: Tuple[ResourceTag, ...]


@dataclass
class JobComponentRequest:
    job_id: bytes
    resource_id: bytes
    component_id: bytes
    config: Optional[ComponentConfig] = None


@dataclass(frozen=True)
class JobMultiComponentRequest:
    job_id: bytes
    resource_id: bytes
    components_allowed: Tuple[bytes, ...] = ()
    components_disallowed: Tuple[bytes, ...] = ()
    tags_ignored: Tuple[ResourceTag, ...] = ()
    all_unpackers: bool = False
    all_identifiers: bool = False
    all_analyzers: bool = False
    all_packers: bool = False



import sys

from keystone import (
    KS_ARCH_ARM64,
    KS_ARCH_ARM,
    KS_ARCH_X86,
    KS_MODE_THUMB,
    KS_MODE_ARM,
    KS_MODE_64,
    KS_MODE_32,
    KS_MODE_16,
    Ks,
    KsError,
    KS_ARCH_PPC,
    KS_MODE_BIG_ENDIAN,
    KS_MODE_LITTLE_ENDIAN,
)

from ofrak.core.architecture import ProgramAttributes
from ofrak.service.assembler.assembler_service_i import AssemblerServiceInterface
from ofrak_io.stream_capture import StreamCapture
from ofrak_type.architecture import InstructionSet, InstructionSetMode, ProcessorType
from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness

X86_64_SPECIAL_CASES = {
    "mov rax, qword ptr fs:[0x28]": b"\x64\x48\x8B\x04\x25\x28\x00\x00\x00",
    "xor rdi, qword ptr fs:[0x28]": b"\x64\x48\x33\x3C\x25\x28\x00\x00\x00",
    "mov rcx, qword ptr fs:[0x28]": b"\x64\x48\x8B\x0C\x25\x28\x00\x00\x00",
    "xor rbx, qword ptr fs:[0x28]": b"\x64\x48\x33\x1C\x25\x28\x00\x00\x00",
    "xor rsi, qword ptr fs:[0x28]": b"\x64\x48\x33\x34\x25\x28\x00\x00\x00",
    "xor rdx, qword ptr fs:[0x28]": b"\x64\x48\x33\x14\x25\x28\x00\x00\x00",
    "xor rcx, qword ptr fs:[0x28]": b"\x64\x48\x33\x0C\x25\x28\x00\x00\x00",
    "xor rax, qword ptr fs:[0x28]": b"\x64\x48\x33\x04\x25\x28\x00\x00\x00",
}


class KeystoneAssemblerService(AssemblerServiceInterface):
    """
    An assembler service implementation using the keystone engine.
    """

    def __init__(self):
        self._ks_by_processor = {}

    @staticmethod
    def _get_keystone_arch_flag(
        program_attributes: ProgramAttributes,
        mode: InstructionSetMode,
    ):
        if program_attributes.isa is InstructionSet.ARM:
            return KS_ARCH_ARM
        elif program_attributes.isa is InstructionSet.AARCH64:
            return KS_ARCH_ARM64
        elif program_attributes.isa is InstructionSet.X86:
            return KS_ARCH_X86
        elif program_attributes.isa is InstructionSet.PPC:
            return KS_ARCH_PPC
        raise ValueError(f"Cannot generate the keystone architecture flag for {program_attributes}")

    @staticmethod
    def _get_keystone_mode_flag(
        program_attributes: ProgramAttributes,
        mode: InstructionSetMode,
    ):
        if program_attributes.endianness is Endianness.BIG_ENDIAN:
            ks_endian_flag = KS_MODE_BIG_ENDIAN
        else:
            ks_endian_flag = KS_MODE_LITTLE_ENDIAN

        if program_attributes.isa is InstructionSet.AARCH64:
            return ks_endian_flag

        elif program_attributes.isa is InstructionSet.ARM:
            mode_flag = 0
            if mode is InstructionSetMode.THUMB:
                # THUMB mode
                return mode_flag | KS_MODE_THUMB | ks_endian_flag
            elif program_attributes.bit_width is BitWidth.BIT_32:
                mode_flag |= KS_MODE_ARM
            if program_attributes.processor is None:
                # Plain ARM 32
                return mode_flag | ks_endian_flag
            if program_attributes.processor is ProcessorType.XSCALE:
                # XSCALE
                raise NotImplementedError("XSCALE not implemented")
            elif program_attributes.processor is ProcessorType.ARM926EJ_S:
                # ARMv5
                raise NotImplementedError("AMRv5 not implemented")
            else:
                # Default 32-bit ARM for other Processor types
                return mode_flag | ks_endian_flag

        elif program_attributes.isa is InstructionSet.X86:
            if program_attributes.bit_width is BitWidth.BIT_64:
                return KS_MODE_64 | ks_endian_flag
            elif program_attributes.bit_width is BitWidth.BIT_32:
                return KS_MODE_32 | ks_endian_flag
            elif program_attributes.bit_width is BitWidth.BIT_16:
                return KS_MODE_16 | ks_endian_flag

        elif program_attributes.isa is InstructionSet.PPC:
            if program_attributes.bit_width == BitWidth.BIT_64:
                return KS_MODE_64 | ks_endian_flag
            elif program_attributes.bit_width == BitWidth.BIT_32:
                return KS_MODE_32 | ks_endian_flag

        raise ValueError(f"Cannot generate the keystone mode flag for {program_attributes}")

    def _get_keystone_instance(
        self,
        program_attributes: ProgramAttributes,
        mode: InstructionSetMode = InstructionSetMode.NONE,
    ) -> Ks:
        """
        Get or build a Keystone instance for the provided processor
        :param program_attributes:
        :param mode:
        """
        ks = self._ks_by_processor.get((program_attributes, mode), None)
        if ks is None:
            ks = Ks(
                self._get_keystone_arch_flag(program_attributes, mode),
                self._get_keystone_mode_flag(program_attributes, mode),
            )
            self._ks_by_processor[(program_attributes, mode)] = ks
        return ks

    async def assemble(
        self,
        assembly: str,
        vm_addr: int,
        program_attributes: ProgramAttributes,
        mode: InstructionSetMode = InstructionSetMode.NONE,
    ) -> bytes:
        """
        Assemble the given assembly code using keystone.

        :param assembly:
        :param vm_addr:
        :param program_attributes:
        :param mode:

        :return: machine code
        """
        # TODO: This is a very temporary fix to T395.
        # TODO: Figure out where to actaully handle situations like this
        assembly_parts = None
        bad_instruction = None
        for instruction in X86_64_SPECIAL_CASES.keys():
            if instruction in assembly:
                assembly_parts = assembly.split(instruction)
                bad_instruction = instruction
                break
        if assembly_parts is not None and bad_instruction is not None:
            machine_code_parts = []
            assembly_size = 0
            for assembly_part in assembly_parts:
                if assembly_part == "":
                    machine_code_parts.append(b"")
                else:
                    machine_code_part = await self.assemble(
                        assembly_part, vm_addr + assembly_size, program_attributes
                    )
                    machine_code_parts.append(machine_code_part)
                    assembly_size += len(machine_code_part)
                assembly_size += 9
            machine_code = X86_64_SPECIAL_CASES[bad_instruction].join(machine_code_parts)
            if machine_code == "":
                return bytes(X86_64_SPECIAL_CASES[bad_instruction])
            return machine_code

        # special register prefix preprocessing for PPC
        preprocessed_assembly = assembly
        if program_attributes.isa is InstructionSet.PPC:
            for prefix in ["r", "f", "v"]:
                for n in range(32):
                    register_operand = f"{prefix}{n}"
                    preprocessed_assembly = preprocessed_assembly.replace(
                        r" %s," % register_operand, " %u," % n
                    )
                    preprocessed_assembly = preprocessed_assembly.replace(
                        r"(%s)" % register_operand, "(%u)" % n
                    )
                    preprocessed_assembly = preprocessed_assembly.replace(
                        r" %s" % register_operand, " %u" % n
                    )

        try:
            ks = self._get_keystone_instance(program_attributes, mode)
            if program_attributes.isa in (InstructionSet.ARM, InstructionSet.AARCH64):
                # This place is a message... and part of a system of messages ...pay attention to it!
                # Sending this message was important to us. We considered ourselves to be a powerful culture.
                # This place is not a place of honor ... no highly esteemed deed is commemorated here... nothing valued is here.
                # What is here was dangerous and repulsive to us. This message is a warning about danger.
                # The danger is in a particular location... it increases towards a center... the center of danger is here... of a particular size and shape, and below us.
                # The danger is still present, in your time, as it was in ours.
                # The danger is to the body, and it can kill.
                # The form of the danger is an emanation of energy.
                # The danger is unleashed only if you substantially disturb this place. This place is best shunned and left uninhabited.
                # Check for warnings in Keystone standard error.
                # If they appear, reset the Ks objects, as bugs in Keystone error handling
                # sometimes cause segmentation faults at subsequent calls to ks.asm.
                # See T403.
                with StreamCapture(sys.stderr) as stream_capture:
                    machine_code, _ = ks.asm(preprocessed_assembly, addr=vm_addr, as_bytes=True)
                if "warning:" in stream_capture.get_captured_stream():
                    self._ks_by_processor = {}
            else:
                machine_code, _ = ks.asm(preprocessed_assembly, addr=vm_addr, as_bytes=True)
            return machine_code
        except KsError as error:

            assembly_vm_addr = vm_addr
            failing_instruction = None

            for assembly_line in preprocessed_assembly.splitlines():
                try:
                    ks = self._get_keystone_instance(program_attributes, mode)
                    machine_code, _ = ks.asm(assembly_line, addr=assembly_vm_addr, as_bytes=True)
                    assembly_vm_addr += len(machine_code)

                except KsError:
                    failing_instruction = assembly_line
                    break

            raise Exception(
                "Keystone ERROR in {}:\n[0x{:x}]\n{}\nerror on instruction '{}' @ 0x{:x}: {}".format(
                    program_attributes.isa,
                    vm_addr,
                    preprocessed_assembly,
                    failing_instruction,
                    assembly_vm_addr,
                    error,
                )
            )



import functools
from typing import Iterable, List, Set, Tuple


class ResourceTag(type):
    all_tags: Set["ResourceTag"] = set()

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        specificity = 0
        for base in cls.base_tags():
            specificity = max(specificity, base.tag_specificity())

        cls._specificity: int = specificity + 1
        ResourceTag.all_tags.add(cls)

    def tag_specificity(cls) -> int:
        """
        Indicates how specific an abstraction this tag is.
        :return: The number of classes in the inheritance hierarchy between this class and
        Resource
        """
        return cls._specificity

    @functools.lru_cache(None)
    def tag_classes(cls) -> Set["ResourceTag"]:
        """
        :return: Set of parent classes (including itself) that are a subclass of a _ResourceTag but
        not the _ResourceTag class.
        """
        parents = set()
        parents.add(cls)
        for base in cls.base_tags():
            parents.update(base.tag_classes())
        return parents

    @functools.lru_cache(None)
    def base_tags(cls) -> Set["ResourceTag"]:
        """
        :return: All _ResourceTags which this tag inherits from directly (does not traverse all
        ancestors)
        """
        base_tags = set()
        for base in cls.__bases__:
            if base is not cls and isinstance(base, ResourceTag) and base.tag_specificity() > 0:
                base_tags.add(base)
        return base_tags

    @staticmethod
    def sort_tags_into_tiers(
        tags: "Iterable[ResourceTag]",
    ) -> "Tuple[Tuple[ResourceTag, ...], ...]":
        """
        Sort the given tags by specificity, and group all the ties together.

        :param tags: Tags to sort and group

        :return: Tuple of groups of tags with the same specificity, sorting all of these by the
        specificity value each group represents from least to greatest.
        """
        levels: List[List[ResourceTag]] = [[], [], [], [], [], [], [], [], [], []]
        for t in tags:
            spec = t.tag_specificity()
            if spec > len(levels):
                levels.extend([] for _ in range(spec - len(levels)))
            levels[spec].append(t)

        return tuple(tuple(level) for level in reversed(levels) if len(level) > 0)

    @classmethod  # pragma: no cover
    def caption(cls, attributes) -> str:  # pragma: no cover
        return str(cls.__name__)

    def __repr__(cls):
        return cls.__name__



import logging
from abc import ABC, abstractmethod
from typing import Optional, Tuple

from ofrak.component.abstract import AbstractComponent
from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.modifier import Modifier
from ofrak.model.component_model import CC
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.component_locator_i import (
    ComponentLocatorInterface,
)
from ofrak.model.component_filters import (
    ComponentWhitelistFilter,
    ComponentTypeFilter,
    ComponentOrMetaFilter,
    ComponentAndMetaFilter,
    ComponentNotMetaFilter,
)
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface

LOGGER = logging.getLogger(__name__)


class PackerError(RuntimeError):
    """Base exception raised by Packers."""


class Packer(AbstractComponent[CC], ABC):
    """
    Packers are components that typically mirror unpackers, taking constituent children resources
    (and sometimes descendants) and reassembling them to produce a new resource.
    """

    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
        component_locator: ComponentLocatorInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self._component_locator = component_locator

    @abstractmethod
    async def pack(self, resource: Resource, config: CC) -> None:
        """
        Pack the given resource.

        Users should not call this method directly; rather, they should run
        [Resource.run][ofrak.resource.Resource.run] or
        [Resource.pack][ofrak.resource.Resource.pack].

        :param resource:
        :param config: Optional config for packing. If an implementation provides a default,
        this default will always be used when config would otherwise be None. Note that a copy of
        the default config will be passed, so the default config values cannot be modified
        persistently by a component run.
        """
        raise NotImplementedError()

    @classmethod
    def get_default_config(cls) -> Optional[CC]:
        return cls._get_default_config_from_method(cls.pack)

    async def _run(self, resource: Resource, config: CC) -> None:
        if resource.has_component_run(self.get_id(), self.get_version()):
            LOGGER.warning(
                f"The {self.get_id().decode()} packer has already been run on resource"
                f" {resource.get_id().hex()}"
            )
            return
        await self.pack(resource, config)
        resource.add_component(self.get_id(), self.get_version())
        # Identify which unpackers ran (if any) and clear that record, so that it will be allowed
        # to run again
        unpacker_ids = self._get_which_unpackers_ran(resource)
        for unpacker_id in unpacker_ids:
            resource.remove_component(unpacker_id)
        for child_r in await resource.get_children():
            await child_r.delete()

    def _get_which_unpackers_ran(self, resource: Resource) -> Tuple[bytes, ...]:
        unpackers_ran = self._component_locator.get_components_matching_filter(
            ComponentAndMetaFilter(
                ComponentWhitelistFilter(*resource.get_model().component_versions.keys()),
                # Use process of elimination to avoid circular import between unpacker.py, packer.py
                ComponentNotMetaFilter(
                    ComponentOrMetaFilter(
                        ComponentTypeFilter(Packer),  # type: ignore
                        ComponentTypeFilter(Analyzer),  # type: ignore
                        ComponentTypeFilter(Identifier),  # type: ignore
                        ComponentTypeFilter(Modifier),  # type: ignore
                    )
                ),
            )
        )
        return tuple(unpacker.get_id() for unpacker in unpackers_ran)



import re
from dataclasses import dataclass

from ofrak.component.modifier import Modifier, ModifierError
from ofrak.core.binary import BinaryPatchConfig, BinaryPatchModifier, GenericText, GenericBinary
from ofrak.core.filesystem import File
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource


@dataclass
class StringPatchingConfig(ComponentConfig):
    """
    Dataclass required to apply a string patch with `StringPatchingModifier`. The configuration
    describes the `offset` where the patch is to be applied, and the `string` to patch in.

    :var offset: the offset at which to apply the patch
    :var string: the string to patch in
    """

    offset: int
    string: str
    null_terminate: bool = False


class StringPatchingModifier(Modifier[StringPatchingConfig]):
    """
    Patch a string in a resource at a given offset, based on the provided configuration.
    """

    id = b"StringPatchingModifier"
    targets = (GenericText,)

    async def modify(self, resource: Resource, config: StringPatchingConfig):
        new_data = config.string.encode("utf-8")
        if config.null_terminate:
            new_data += b"\x00"
        patch_config = BinaryPatchConfig(config.offset, new_data)
        await resource.run(BinaryPatchModifier, patch_config)


@dataclass
class StringFindReplaceConfig(ComponentConfig):
    """
    :var to_find: the string to search for
    :var replace_with: the string to pass in
    :var null_terminate: add a null terminator to the replacement if True
    :var allow_overflow: allow the replace string to overflow the found string if True
    """

    to_find: str
    replace_with: str
    null_terminate: bool = True
    allow_overflow: bool = False


class StringFindReplaceModifier(Modifier[StringFindReplaceConfig]):
    """
    Find and replace all instances of a given string with a replacement string.
    """

    targets = (GenericBinary, File)

    async def modify(self, resource: Resource, config: StringFindReplaceConfig) -> None:
        to_find = config.to_find.encode("utf-8")
        replace_with = config.replace_with.encode("utf-8") + (
            b"\x00" if config.null_terminate and config.replace_with[-1] != "\x00" else b""
        )
        if not config.allow_overflow and len(replace_with) > len(to_find):
            raise ModifierError(
                f"Original string is longer than the new string ({len(to_find)} < "
                f"{len(replace_with)})! Set config.allow_overflow = True to override this error. "
                f"If you expect that the string to replace is null-terminated, then an overflow "
                f"of one byte when config.null_terminate = True will not have any effect."
            )
        original_data = await resource.get_data()
        offsets = [m.start() for m in re.finditer(to_find, original_data)]
        for offset in offsets:
            await resource.run(BinaryPatchModifier, BinaryPatchConfig(offset, replace_with))



from typing import Any, List, Dict

import orjson
from inspect import isfunction

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface
from ofrak.service.serialization.service_i import SerializationServiceInterface


class PJSONSerializationService(SerializationServiceInterface):
    """
    Service handling the serialization and deserialization of various types into a JSON-compatible Python type
    named PJSON (for "proto-JSON"). The idea is that this is a Python type, making its handling easier, but which has a
    straightforward mapping to real JSON via json.dump() or json.load().

    In this model, JSON serialization happens in two stages: object to PJSON, then PJSON to JSON (string).
    Deserialization works the same way: JSON (string) to PJSON, then PJSON to object.

    Note that the PJSON representation of an object might be the object itself, for example to_pjson(1, int) == 1.
    """

    def __init__(self, serializers: List[SerializerInterface]):
        """
        This class should be instantiated via the dependency injector. This assumes that the `serializers`
        module has been discovered.
        """
        self._serializers = serializers
        for serializer in self._serializers:
            # This class requires the custom serializers, and the custom serializers require this class.
            # Hacky workaround for this circular dependency.
            setattr(serializer, "_service", self)

        # Used to cache results of serializer discovery
        self._cached_type_to_serializer_mapping: Dict[Any, SerializerInterface] = {}

    def to_pjson(self, obj: Any, type_hint: Any) -> PJSONType:
        """
        Generic function to recursively convert `obj` into PJSON.

        The following type hints are supported:

        - Tuple, List, Sequence, Iterable
        - Dict
        - Set
        - Union, and its special case Optional
        (NOTE: Union should only be used with mutually exclusive types, and types with different PJSON representations.
        For example, serializing bytes as Union[str, bytes] will work, but then since both str and bytes are serialized as
        str, deserialization won't be able to figure out whether the PJSON str object should be decoded as str or bytes.
        In practice, all types within the Union are tried in order, and the first that works is selected. So serializing
        then deserializing bytes as Union[str, bytes] will result in a str, not the original bytes. This should be
        considered an implementation detail, though: relying on the order of the types within the Union isn't encouraged.)
        - int, float, bool, str
        - type(None)
        - bytes
        - enum.Enum, or any type hint corresponding to a subclass of enum.Enum
        - Type[X], type or another metaclass, when `obj` is a class and should be serialized as a reference to itself
        - dataclass instances
        - other class instances. This will work if the class has type annotations which correspond to the arguments in its
        __init__() method, and the attributes don't change meaningfully as they go through __init__() (i.e. the class
        isn't too different from a dataclass).
        - Any, only if the object is already of PJSONType. Otherwise TypeError is raised. (prefer specific types whenever
        possible)

        If any type is found during recursion that isn't in this list, a TypeError exception is raised.
        """
        serializer = self._get_serializer(obj, type_hint)
        return serializer.obj_to_pjson(obj, type_hint)

    def from_pjson(self, pjson_obj: PJSONType, type_hint: Any) -> Any:
        """Opposite of `to_pjson`."""
        serializer = self._get_serializer(pjson_obj, type_hint)
        return serializer.pjson_to_obj(pjson_obj, type_hint)

    def _get_serializer(self, obj: Any, type_hint: Any) -> SerializerInterface:
        """Return the first serializer/deserializer pair found for `type_hint`."""
        # Has this type already been seen?
        try:
            return self._cached_type_to_serializer_mapping[type_hint]
        except KeyError:
            pass
        # First pass: `targets` as explicit types have priority over predicates
        for serializer in self._serializers:
            for target in serializer.targets:
                if not isfunction(target) and target == type_hint:
                    self._cached_type_to_serializer_mapping[type_hint] = serializer
                    return serializer
        # Second pass: if the type hint didn't correspond to any explicit type, try predicates
        for serializer in self._serializers:
            for target in serializer.targets:
                if isfunction(target) and target(type_hint) is True:
                    self._cached_type_to_serializer_mapping[type_hint] = serializer
                    return serializer
        raise TypeError(f"Unrecognized type hint {type_hint} for {obj}")

    def dumps(self, pjson_obj: PJSONType) -> str:
        """Wrapper around the dumping method of the JSON library used."""
        return orjson.dumps(pjson_obj).decode("utf-8")

    def loads(self, json_obj: str) -> PJSONType:
        """Wrapper around the loading method of the JSON library used."""
        return orjson.loads(bytes(json_obj, "utf-8"))

    def to_json(self, obj: Any, type_hint: Any) -> str:
        return self.dumps(self.to_pjson(obj, type_hint))

    def from_json(self, json_obj: str, type_hint: Any) -> Any:
        return self.from_pjson(self.loads(json_obj), type_hint)



from abc import ABC, abstractmethod

from ofrak import OFRAKContext
from ofrak.resource import Resource


class ModifyPattern(ABC):
    """
    Generic test pattern for modification. The test expects to:
    1. Create an initial root resource
    2. Modify the initial root resource
    3. Verify that the root resource matches some expected pattern

    Each step is broken out into an abstractmethod which subclasses of this pattern should
    implement.
    """

    async def test_modify(self, ofrak_context):
        root_resource = await self.create_root_resource(ofrak_context)
        await self.modify(root_resource)
        await self.verify(root_resource)

    @abstractmethod
    async def create_root_resource(self, ofrak_context: OFRAKContext) -> Resource:
        """
        Create and return the root resource which will be unpacked
        """
        raise NotImplementedError()

    @abstractmethod
    async def modify(self, root_resource: Resource) -> None:
        """
        Modify the root resource
        """
        raise NotImplementedError()

    @abstractmethod
    async def verify(self, root_resource: Resource) -> None:
        """
        Verify that the root resource matches what is expected after repacking with the
        modifications
        """
        raise NotImplementedError()



from typing import Any, Dict, Set

from ofrak.model.data_model import DataModel
from ofrak.service.data_service import DataService, _DataRoot, DataId, _GridXAxisT
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class DataRootSerializer(SerializerInterface):

    targets = (_DataRoot,)

    data_root_annotations = {
        "model": DataModel,
        "data": bytes,
        "_children": Dict[DataId, DataModel],
    }

    def obj_to_pjson(self, obj: _DataRoot, type_hint: Any) -> Dict[str, PJSONType]:
        data_root_pjson = {
            attr_name: self._service.to_pjson(getattr(obj, attr_name), type_hint)
            for attr_name, type_hint in self.data_root_annotations.items()
        }
        return data_root_pjson

    def pjson_to_obj(self, pjson_obj: Dict[str, PJSONType], type_hint: Any) -> _DataRoot:
        root_model = self._service.from_pjson(pjson_obj["model"], DataModel)
        data = self._service.from_pjson(pjson_obj["data"], bytes)
        children = self._service.from_pjson(pjson_obj["_children"], Dict[DataId, DataModel])

        grid_starts_first: _GridXAxisT = []
        grid_ends_first: _GridXAxisT = []

        for model in children.values():
            default_set: Set[DataId] = set()
            set1 = _DataRoot._add_to_grid(
                grid_starts_first, model.range.start, model.range.end, default_set
            )
            set2 = _DataRoot._add_to_grid(
                grid_ends_first, model.range.end, model.range.start, default_set
            )

            assert set1 is set2  # These sets should always be the same!
            set1.add(model.id)

        data_root = _DataRoot(root_model, data)
        data_root._children = children
        data_root._grid_starts_first = grid_starts_first
        data_root._grid_ends_first = grid_ends_first
        return data_root


class DataServiceSerializer(SerializerInterface):
    """
    Serialize and deserialize `DataService` into `PJSONType`.

    The `DataService` has some internal data structures which require custom
    serialization/deserialization, specifically the _DataRoot and _Waypoint. This is done with
    methods of this class, and avoids serializing redundant information.
    """

    targets = (DataService,)

    def obj_to_pjson(self, obj: DataService, _type_hint: Any) -> PJSONType:
        return self._service.to_pjson(getattr(obj, "_roots"), Dict[DataId, _DataRoot])

    def pjson_to_obj(self, pjson_obj: Dict[str, Dict], _type_hint: Any) -> DataService:
        model_store = {}
        roots = self._service.from_pjson(pjson_obj, Dict[DataId, _DataRoot])
        # The DataService assumes that the DataModel instance in the model store is the same
        # instance in the _DataRoot.model (and _DataRoot._children). Ensure that this is the case
        # before returning the DataService.
        for data_id, root_model in roots.items():
            model_store[data_id] = root_model.model
            for child in root_model.get_children():
                model_store[child.id] = child
        data_service: DataService = DataService.__new__(DataService)
        data_service._model_store = model_store
        data_service._roots = roots
        return data_service



import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Tuple, Iterable

from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource

from ofrak.component.analyzer import Analyzer
from ofrak.component.unpacker import Unpacker
from ofrak.core.addressable import Addressable
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.basic_block import BasicBlock
from ofrak.core.data import DataWord
from ofrak.core.memory_region import MemoryRegion
from ofrak.model.resource_model import ResourceAttributes, index
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class ComplexBlockDataReferenceAttributes(ResourceAttributes):
    referenced_data_vm_addresses: Tuple[int, ...]


@dataclass
class ComplexBlock(MemoryRegion):
    """
    A collection of [basic blocks][ofrak.core.basic_block.BasicBlock] and
    [data words][ofrak.core.data.DataWord] that represent a logical unit of code (usually a
    function).

    :ivar virtual_address: the lowest virtual address in the complex block
    :ivar size: the size of the complex block
    :ivar name: the complex block's name
    """

    name: str

    @index
    def Symbol(self) -> str:
        return self.name

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            cb_attributes = all_attributes[AttributesType[ComplexBlock]]
            addressable_attributes = all_attributes[AttributesType[Addressable]]
        except KeyError:
            return super().caption(all_attributes)
        return f"{hex(addressable_attributes.virtual_address)}: {cb_attributes.name}"

    async def get_basic_blocks(self) -> Iterable[BasicBlock]:
        """
        Get complex block's basic blocks.

        :return: basic blocks
        """
        return await self.resource.get_descendants_as_view(
            BasicBlock,
            r_filter=ResourceFilter.with_tags(BasicBlock),
            r_sort=ResourceSort(Addressable.VirtualAddress),
        )

    async def get_assembly(self) -> str:
        """
        Get the complex block's instructions as an assembly string.

        :return: the complex block's assembly
        """
        bbs = await self.get_basic_blocks()
        bb_assemblies = [bb_r.get_assembly() for bb_r in bbs]
        return "\n".join(await asyncio.gather(*bb_assemblies))

    async def get_data_words(self) -> Iterable[DataWord]:
        """
        Get the complex block's [data words][ofrak.core.data.DataWord].

        :return: the data words in the complex block
        """
        return await self.resource.get_descendants_as_view(
            DataWord,
            r_filter=ResourceFilter.with_tags(DataWord),
            r_sort=ResourceSort(Addressable.VirtualAddress),
        )

    async def get_mode(self) -> InstructionSetMode:
        """
        Get the complex block's [mode][ofrak_type.architecture.InstructionSetMode].

        :raises ValueError: if the basic blocks in the complex block have more than one mode
        :return: the mode of the complex block
        """
        await self.resource.unpack()
        bb_modes = {bb.mode for bb in await self.get_basic_blocks()}
        if len(bb_modes) == 1:
            return bb_modes.pop()
        elif len(bb_modes) > 1:
            raise ValueError(
                f"Multiple modes present in complex block! Not all basic blocks have "
                f"the same mode; found modes {bb_modes}"
            )
        else:
            raise ValueError("No basic blocks found in complex block! Perhaps it was not unpacked")


class ComplexBlockStructureError(RuntimeError):
    pass


class ComplexBlockUnpacker(Unpacker[None], ABC):
    """
    Unpack a [complex block][ofrak.core.complex_block.ComplexBlock] into
    [basic blocks][ofrak.core.basic_block.BasicBlock] and [data words][ofrak.core.data.DataWord].
    """

    targets = (ComplexBlock,)
    children = (BasicBlock, DataWord)
    id = b"ComplexBlockUnpacker"

    @abstractmethod
    async def unpack(self, resource: Resource, config=None):
        """
        Unpack a complex block, identifying all of the basic blocks and data words which are a part
        of it.

        The identified basic blocks and data words must be within the previously identified range
        of the complex block. If the analysis engine identifies basic blocks outside of this
        range, those are be ignored - i.e. not unpacked - and the rest of the basic blocks in
        the function are unpacked as usual.
        """
        raise NotImplementedError()


class ComplexBlockAnalyzer(Analyzer[None, ComplexBlock], ABC):
    """
    Analyze a [complex block][ofrak.core.complex_block.ComplexBlock] and extract its virtual
    address, size, and name.
    """

    targets = (ComplexBlock,)
    outputs = (ComplexBlock,)
    id = b"ComplexBlockAnalyzer"

    @abstractmethod
    async def analyze(self, resource: Resource, config=None) -> ComplexBlock:
        """
        Analyze a complex block resource and extract its virtual address, size, and name.

        :param resource: the complex block resource
        :param config:

        :return: the analyzed complex block
        """
        raise NotImplementedError()


class DataRefsAnalyzer(Analyzer[None, Tuple[ComplexBlockDataReferenceAttributes]], ABC):
    """
    Analyze the references a [complex block][ofrak.core.complex_block.ComplexBlock] makes to data
    addresses.
    """

    id = b"DataRefsAnalyzer"
    targets = (ComplexBlock,)
    outputs = (ComplexBlockDataReferenceAttributes,)

    @abstractmethod
    async def analyze(
        self, resource: Resource, config=None
    ) -> Tuple[ComplexBlockDataReferenceAttributes]:
        """
        Analyze the references a complex block resource makes to data addresses

        :param resource: the complex block resource
        :param config:

        :return: The virtual addresses of all the data words referenced by the complex block.
        """
        raise NotImplementedError()



"""
Assembler services used to assemble and disassemble code.
"""
from abc import ABCMeta, abstractmethod

from ofrak.core.architecture import ProgramAttributes
from ofrak_type.architecture import InstructionSetMode
from ofrak.service.abstract_ofrak_service import AbstractOfrakService


class AssemblerServiceInterface(AbstractOfrakService):
    """An interface for assembler services."""

    __metaclass__ = ABCMeta

    @abstractmethod
    async def assemble(
        self,
        assembly: str,
        vm_addr: int,
        program_attributes: ProgramAttributes,
        mode: InstructionSetMode = InstructionSetMode.NONE,
    ) -> bytes:
        """
        Assemble the given assembly code.

        :param str assembly: The assembly to assemble
        :param int vm_addr: The virtual address at which the assembly should be assembled.
        :param ProgramAttributes program_attributes: The processor targeted by the assembly
        :param InstructionSetMode mode: The mode of the processor for the assembly

        :return: The assembled machine code
        """
        raise NotImplementedError



import bz2
import logging
from dataclasses import dataclass

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicDescriptionIdentifier, MagicMimeIdentifier
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)


@dataclass
class Bzip2Data(GenericBinary):
    """
    A bzip2 binary blob.
    """


class Bzip2Unpacker(Unpacker[None]):
    """
    Unpack bzip2 data.

    bzip2 binary blobs decompress into one child.
    """

    targets = (Bzip2Data,)
    children = (GenericBinary,)

    async def unpack(self, resource: Resource, config=None):
        """
        Unpack bzip2 data.

        :param resource:
        :param config:
        """
        resource_data = await resource.get_data()
        decompressed_data = bz2.decompress(resource_data)
        await resource.create_child(
            tags=(GenericBinary,),
            data=decompressed_data,
        )


class Bzip2Packer(Packer[None]):
    """
    Pack a resource, compressing it into bzip2 data.
    """

    targets = (Bzip2Data,)

    async def pack(self, resource: Resource, config=None):
        """
        Pack a resource into bzip2 data.

        :param resource:
        :param config:
        """
        bzip2_child = await resource.get_only_child()
        bzip2_compressed = bz2.compress(await bzip2_child.get_data())
        original_size = await resource.get_data_length()
        resource.queue_patch(Range(0, original_size), bzip2_compressed)


MagicMimeIdentifier.register(Bzip2Data, "application/x-bzip2")
MagicDescriptionIdentifier.register(Bzip2Data, lambda s: s.startswith("BZip2 archive"))



from abc import ABC, abstractmethod
from difflib import ndiff
from subprocess import CalledProcessError

from ofrak import OFRAKContext
from ofrak.component.abstract import ComponentSubprocessError
from ofrak.resource import Resource


class UnpackModifyPackPattern(ABC):
    """
    Generic test pattern for unpacker/packer pairing. The test expects to:
    1. Create an initial root resource
    2. Unpack that root resource
    3. Get one of the newly-created descendants of the root resource and modify that descendant
    4. Repack the root resource - because of the modification in step 4, the root resource should
    now have different data after this step.
    5. Verify that the repacked root resource matches some expected pattern

    Each step is broken out into an abstractmethod which subclasses of this pattern should
    implement.
    """

    async def test_unpack_modify_pack(self, ofrak_context):
        try:
            root_resource = await self.create_root_resource(ofrak_context)
            await self.unpack(root_resource)
            original_tree = await root_resource.summarize_tree()
            await self.modify(root_resource)
            modified_tree = await root_resource.summarize_tree()
            print("\n")
            print(
                "".join(
                    ndiff(
                        original_tree.splitlines(keepends=True),
                        modified_tree.splitlines(keepends=True),
                    )
                ),
                end="",
            )
            await self.repack(root_resource)
            await self.verify(root_resource)
        except CalledProcessError as e:
            # Better printing of errors if something goes wrong in test setup/execution
            raise ComponentSubprocessError(e)

    @abstractmethod
    async def create_root_resource(self, ofrak_context: OFRAKContext) -> Resource:
        """
        Create and return the root resource which will be unpacked
        """
        raise NotImplementedError()

    @abstractmethod
    async def unpack(self, root_resource: Resource) -> None:
        """
        Unpack the root resource
        """
        raise NotImplementedError()

    @abstractmethod
    async def modify(self, unpacked_root_resource: Resource) -> None:
        """
        Once the root resource is unpacked, the test needs to modify at least one of its
        descendants. This method should contain the logic to get those resource(s).
        """
        raise NotImplementedError()

    @abstractmethod
    async def repack(self, modified_root_resource: Resource) -> None:
        """
        Pack the root resource after it has been modified
        """
        raise NotImplementedError()

    @abstractmethod
    async def verify(self, repacked_root_resource: Resource) -> None:
        """
        Verify that the repacked root resource matches what is expected after repacking with the
        modifications
        """
        raise NotImplementedError()


class UnpackPackPattern(UnpackModifyPackPattern, ABC):
    async def modify(self, unpacked_root_resource: Resource) -> None:
        pass



from typing import Dict, Any, List, Tuple, Type, Set, Optional, Union

# Under-approximation of the PJSON type (which is recursive, contrary to this definition).
# All objects of the real PJSON type will be of this type, but some non-PJSON objects will
# also be of this type.
PJSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any], Tuple[Any, ...]]

# The containers we descend into
ContainerType = Union[
    Optional[Any], Dict[Any, Any], List[Any], Tuple[Any, ...], Set[Any], Type[Any]
]



import logging
import lzma
from io import BytesIO
from typing import Union

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)


class LzmaData(GenericBinary):
    """
    An lzma binary blob.
    """

    async def get_child(self) -> GenericBinary:
        return await self.resource.get_only_child_as_view(GenericBinary)


class XzData(GenericBinary):
    """
    An xz binary blob.
    """

    async def get_child(self) -> GenericBinary:
        return await self.resource.get_only_child_as_view(GenericBinary)


class LzmaUnpacker(Unpacker[None]):
    """
    Unpack (decompress) an LZMA | XZ file.
    """

    id = b"LzmaUnpacker"
    targets = (LzmaData, XzData)
    children = (GenericBinary,)

    async def unpack(self, resource: Resource, config=None):
        file_data = BytesIO(await resource.get_data())

        format = lzma.FORMAT_AUTO

        if resource.has_tag(XzData):
            format = lzma.FORMAT_XZ
        elif resource.has_tag(LzmaData):
            format = lzma.FORMAT_ALONE

        lzma_entry_data = None
        compressed_data = file_data.read()

        try:
            lzma_entry_data = lzma.decompress(compressed_data, format)
        except lzma.LZMAError:
            LOGGER.info("Initial LZMA decompression failed. Trying with null bytes stripped")
            lzma_entry_data = lzma.decompress(compressed_data.rstrip(b"\x00"), format)

        if lzma_entry_data is not None:
            await resource.create_child(
                tags=(GenericBinary,),
                data=lzma_entry_data,
            )
        else:
            raise lzma.LZMAError("Decompressed LZMA data is null")


class LzmaPacker(Packer[None]):
    """
    Pack data into a compressed LZMA | XZ file.
    """

    targets = (LzmaData, XzData)

    async def pack(self, resource: Resource, config=None):
        lzma_format, tag = await self._get_lzma_format_and_tag(resource)
        lzma_file: Union[XzData, LzmaData] = await resource.view_as(tag)

        lzma_child = await lzma_file.get_child()
        lzma_compressed = lzma.compress(await lzma_child.resource.get_data(), lzma_format)

        original_size = await lzma_file.resource.get_data_length()
        resource.queue_patch(Range(0, original_size), lzma_compressed)

    async def _get_lzma_format_and_tag(self, resource):
        if resource.has_tag(XzData):
            tag = XzData
            lzma_format = lzma.FORMAT_XZ
        elif resource.has_tag(LzmaData):
            tag = LzmaData
            lzma_format = lzma.FORMAT_ALONE
        else:
            raise TypeError(
                f"Expected target of {self.get_id().decode()} to be either XzFile or LzmaFile"
            )
        return lzma_format, tag


MagicMimeIdentifier.register(LzmaData, "application/x-lzma")
MagicMimeIdentifier.register(XzData, "application/x-xz")
MagicDescriptionIdentifier.register(LzmaData, lambda s: s.startswith("LZMA compressed data"))
MagicDescriptionIdentifier.register(XzData, lambda s: s.startswith("XZ compressed data"))



from typing import Tuple, Optional

from ofrak.service.resource_service_i import ResourceFilter

from ofrak.resource import Resource

from ofrak.component.analyzer import Analyzer

from ofrak.core.architecture import ProgramAttributes
from ofrak.core.memory_region import MemoryRegion
from ofrak.core.program import Program
from ofrak.model.component_model import ComponentConfig


class MemoryRegionProgramAttributesAnalyzer(Analyzer[None, Tuple[ProgramAttributes]]):
    """
    Analyze a [MemoryRegion][ofrak.core.memory_region.MemoryRegion] to extract its
    [ProgramAttributes][ofrak.core.architecture.ProgramAttributes].
    """

    targets = (MemoryRegion,)
    outputs = (ProgramAttributes,)

    async def analyze(
        self, resource: Resource, config: Optional[ComponentConfig] = None
    ) -> Tuple[ProgramAttributes]:
        program_r = await resource.get_only_ancestor(ResourceFilter.with_tags(Program))
        program_attrs = await program_r.analyze(ProgramAttributes)
        return (program_attrs,)



from typing import Any, Type, Dict

from typing_inspect import get_origin

from ofrak.model.resource_model import ResourceIndexedAttribute, ResourceAttributes
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


def _is_generic_resource_indexed_attribute(type_hint):
    return get_origin(type_hint) == ResourceIndexedAttribute


class ResourceIndexedAttributeSerializer(SerializerInterface):
    """
    Serialize and deserialize `ResourceIndexedAttribute` and `ResourceIndexedAttribute[T]` into `PJSONType`.

    A ResourceIndexedAttributes instance can contain a Callable, which is hard to serialize.
    However, we can recover this ResourceIndexedAttributes instance if we know its
    attributes_owner and its name, both being easy to serialize.
    """

    targets = (ResourceIndexedAttribute, _is_generic_resource_indexed_attribute)

    def obj_to_pjson(self, obj: Any, type_hint: Any) -> Dict[str, PJSONType]:
        if obj.attributes_owner is None:
            raise TypeError(f"Cannot serialize {obj.__name__} because its owner has not been set.")
        else:
            return {
                "attributes_owner": self._service.to_pjson(
                    obj.attributes_owner, Type[ResourceAttributes]
                ),
                "name": obj.index_name,
            }

    def pjson_to_obj(self, pjson_obj: Dict[str, PJSONType], type_hint: Any) -> "ResourceIndexedAttribute":  # type: ignore
        attributes_owner = self._service.from_pjson(
            pjson_obj["attributes_owner"], Type[ResourceAttributes]
        )
        name = self._service.from_pjson(pjson_obj["name"], str)
        return getattr(attributes_owner, name)
